;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;               PROOF CHECKER FOR DURATION CALCULUS IN PVS
;;; File: dc-bdd.lisp
;;; Date: Sun Oct 12 1997
;;; Original version for PVS1 by:
;;;   Jens Ulrik Skakkebaek and Sam Owre
;;; Adapted for PVS2 by:
;;;   Soeren T. Heilmann, sh@it.dtu.dk
;;;   Dept. of Information Technology, Technical University of Denmark
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'PVS)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Special recognizer formulas used to detect special DC language constructs

;; Returns T if expr is a DC almost-everywhere formula: 'p'
;; otherwise nil
(defun is-dc-pr? (expr) ;;; Assuming NO outermost dc index
  (and (application? expr)
       (let ((op (operator expr)))
	 (and (eq (id (operator expr)) '|pr|)
	      (eq (id (module-instance (resolution op))) '|dc_dc|)))))


;; Returns T if expr is a DC leadsto formula: A --> B
;; otherwise nil
(defun is-dc-leadsto? (expr) ;;; Assuming NO outermost dc index
  (and (application? expr)
       (let ((op (operator expr)))
	 (and (eq (id (operator expr)) '|leadsto|)
	      (eq (id (module-instance (resolution op))) '|dc_dc|)))))


;; Returns T if expr is a DC point formula: ''
;; otherwise nil
(defun is-dc-point? (expr) ;;; Assuming NO outermost dc index
  (and (name-expr? expr)
       (let ((op expr))
	 (and (eq (id expr) '|point|)
	      (eq (id (module-instance (resolution op))) '|dc_dc|)))))


;; Returns T if expr is an equality of the form: a=b and if = is declared in
;; dcp_value_state.pvs, otherwise return nil.
(defun is-dc-value-state-equality? (expr)
  (and (match-pattern? "a_=a_" expr "a_")
       (defined-in? (operator expr) '|dcp_state_equal|)))
  

;; Returns T if expr is a lifting: n(...) from a type to be a function of time
;; otherwise nil
(defun is-dc-value-state-n? (expr)
  (and (match-pattern? "n(a_)" expr "a_")
       (defined-in? (operator expr) '|dcp_state_coercion|)))


;; Returns T if expr has the type State
(defun is-dc-state-application? (expr)
  (and (application? expr)
       (is-type? expr "State[bool]")))


;; Returns T if expr is a DC expression of an enumeration type 
(defun is-dc-enum? (expr) ;;; Assuming NO outermost dc index
  (and (is-function-of-time? expr)
       (dc-enum-adt? (range (find-supertype (type expr))))))



;; Returns T if expr is of an enumeration type 
;;; Renamed from enum-adt? during porting to avoid name clash with PVS2
(defun dc-enum-adt? (expr)
  (and (adt? expr)
       (enumtype? (adt expr))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constructor functions

;; Make a PVS operator with the name "valid", which is used to assert formulas,
;; decided to be valid, in the sequent.
(defun make-dc-valid (exp)
  (let* ((op (pc-parse "valid" 'expr))
        (res (mk-application op (car exp))))
    (typecheck res :expected *boolean*)))


;; Make a PVS equality whith the resulting type being State
(defun make-dc-value-state-equality (var constr)
  (make-dc (format nil "~a=n(~a)"  (id var)  (id constr)) "State[bool]"))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions to convert PVS expressions to BDD expressions


;; Converts PVS expressions of the right syntax to the syntax required by the
;; DC BDD package.
;;; Renamed from convert-pvs-to-bdd* during porting to avoid name clash with
;;; PVS2
(defun dc-convert-pvs-to-bdd* (expr)
  (cond ((is-dc-indexed? expr) (dc-convert-pvs-to-bdd* (operator expr)))
        ((is-dc-point? expr) "''")
        ((is-dc-n? expr) (dc-convert-pvs-to-bdd* (args1 expr)))
        ((is-dc-value-state-n? expr) (dc-convert-pvs-to-bdd* (args1 expr)))
        ((exists-expr? expr) (break))
        ((forall-expr? expr) (break))
        ((is-dc-conj? expr) 
 	     (format nil "(~a)&(~a)" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-disj? expr)
             (format nil "(~a)|(~a)" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-impl? expr)
             (format nil "(~a)=>(~a)" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-iff? expr)
             (format nil "(~a)<=>(~a)" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-chop? expr)
             (format nil "(~a);(~a)" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-neg? expr)
             (format nil "(~~(~a))" (dc-convert-pvs-to-bdd* (args1 expr))))
        ((is-dc-tt? expr) (format nil "(tt)"))
        ((is-dc-ff? expr) (format nil "(ff)"))
        ((is-dc-box? expr)
             (format nil "([](~a))" (dc-convert-pvs-to-bdd* (args1 expr))))
        ((is-dc-diamond? expr)
             (format nil "(<>(~a))" (dc-convert-pvs-to-bdd* (args1 expr))))
        ((is-dc-leadsto? expr) 
 	     (format nil "\"^(leadsto \"~a\" \"~a\")^\"" (dc-convert-pvs-to-bdd* (args1 expr))
		                 (dc-convert-pvs-to-bdd* (args2 expr))))
        ((is-dc-false? expr) "false")
        ((is-dc-true? expr) "true")
        ((is-dc-pr? expr)
             (format nil "'~a'" (dc-convert-pvs-to-bdd* (args1 expr))))
	((is-dc-value-state-equality? expr)
	 (make-dc-bdd-equality-variable expr))
	(t (make-dc-bdd-variable expr))))


;; Given an operator as a string, a base element as a string, and a list of
;; arguments, return a string where the operator is applied recursively to the
;; list of arguments, terminating with the base element as the last argument.
(defun mk-rec-string (op base args)
  (cond ((null args) base)
	((null (cdr args)) (car args))
	(t (format nil "~a ~a ~a" (car args) op
		   (mk-rec-string op base (cdr args))))))


;; Given a sequent, return a string where the formulas on the left-hand-side
;; have been converted to the BDD syntax an are all conjoined.
(defun left-bdd-conj (seq) 
   (let* ((sfs  (select-seq seq '-))
	  (fmls (mapcar #'(lambda (x) (format nil "(~a)" (dc-convert-pvs-to-bdd*
				       (args1 (formula x))))) sfs)))
        (mk-rec-string "&" "true" fmls)))


;; Given a sequent, return a string where the formulas on the right-hand-side
;; have been converted to the BDD syntax an are all disjoined.
(defun right-bdd-disj (seq) 
   (let* ((sfs  (select-seq seq '+))
	  (fmls (mapcar #'(lambda (x) (format nil "(~a)" (dc-convert-pvs-to-bdd*
				       (formula x)))) sfs)))
        (mk-rec-string "|" "false" fmls)))


;; Given a sequent, return a string where the sequent formulas have all been
;; converted to the BBD syntax and the sequent is converted to a logical 
;; formula in BDD syntax with the same meaning as the original sequent.
(defun mk-dc-bdd-string (seq)
    (format nil "(~a) => (~a)" (left-bdd-conj seq)
;;	    (right-bdd-disj (s-forms *goal*))))
	    (right-bdd-disj seq)))



(defun make-pvs-bdd-string (interval seq)
  (make-dc-valid 
   (list (make-dc-rec "=>" "TRUE" "IForm"
	  (list (collect-dc-antec interval seq)
		(collect-dc-conseq interval seq))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Checking whether PVS expressions conforms to the BDD syntax


; Given an expr, make sure it conforms to the syntax required by the DC BDD
; package.  Returns NIL if syntax is OK, otherwise a nonempty list.
(defun check-dc-bdd-formula (expr)
  (cond ((is-dc-point? expr) nil)
        ((or (is-dc-conj? expr) (is-dc-disj? expr) (is-dc-impl? expr)
	    (is-dc-iff? expr) (is-dc-chop? expr) (is-dc-leadsto? expr))
	 (append (check-dc-bdd-formula (args1 expr))
		 (check-dc-bdd-formula (args2 expr))))
	((or (is-dc-neg? expr)(is-dc-box? expr) (is-dc-diamond? expr))
	 (check-dc-bdd-formula (args1 expr)))
	((is-dc-false? expr) nil)
	((is-dc-true? expr) nil)
	((is-dc-pr? expr)
	  (check-dc-bdd-state (args1 expr)))
	(t (list T))))


; Given an expr, make sure it conforms to the syntax of states required by the
; DC BDD package.  Returns NIL if syntax is OK, otherwise a nonempty list.
(defun check-dc-bdd-state (expr)
  (cond ((or (is-dc-conj? expr) (is-dc-disj? expr) (is-dc-impl? expr)
	     (is-dc-iff? expr) )
	 (append (check-dc-bdd-state (args1 expr))
		 (check-dc-bdd-state (args2 expr))))
	((is-dc-neg? expr) (check-dc-bdd-state (args1 expr)))
	((is-dc-false? expr) nil)
	((is-dc-true? expr) nil)
	((name-expr? expr) nil)
	((is-dc-value-state-equality? expr) nil)
	((is-dc-state-application? expr) nil)
	(t (list T))))


;; Select from the goal sequent all the sequent formulas which conform to the
;; BDD syntax
(defun find-all-dc-bdd-fmlas (sformnum nums)
  (find-all-dc-nums sformnum nums #'(lambda(x) (not (check-dc-bdd-formula x)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions to interface to the DC BDD package

;; Given a file name, return a string of the file contents.  
;;; Renamed from get-bdd-output-string during porting to avoid name clash with
;;; PVS2
#+lucid
(defun dc-get-bdd-output-string (filename)
  (let ((tokstrm (make-string-output-stream)))
    (with-open-file (f filename)
      (loop
       (let ((char (fast-read-char f nil)))
	 (unless char (return nil))
	 (format tokstrm "~a" char)
	 )))
    (get-output-stream-string tokstrm)))

#+allegro
(defun dc-get-bdd-output-string (filename)
  (let ((tokstrm (make-string-output-stream)))
    (with-open-file (f filename)
      (loop
       (let ((char (read-char f nil)))
         (unless char (return nil))
         (format tokstrm "~a" char)
         )))
    (get-output-stream-string tokstrm)))
 


;; Given a BDD string, write it to a file, invoke the DC BDD package on the
;; file, and return the string which is passed from the BDD package.
;;; Renamed from run-program during porting to avoid name clash with PVS2
;#+allegro
;(defun run-program (command &key arguments )
;  (let ((string (format nil "setenv PATH ~a;~a ~{~a ~}"
;                  (sys:getenv "PATH") command arguments)))
;    (excl:shell string)))

#+allegro
(setf (symbol-function 'dc-run-program) #'excl:run-shell-command)
 
;;; Works on and allegro 4.3
#+(version>= 4 3)
(defun dc-bdd-output (bdd-string)
  (let* ((bdd-out-tmp-file
	  (write-to-temp-file ""))
         ;; this is just to get the name of the bddoutfile
	 (stdin-tmp-file 
	  (write-to-temp-file (format nil "~a \"~a\"" bdd-string bdd-out-tmp-file)))
	 (stdout-tmp-file
	  (write-to-temp-file ""))
         ;; this is just to get the name of the stdout-tmp-file
;	 (stderr-tmp-file
;	  (write-to-temp-file ""))
         ;; this is just to get the name of the stderr-tmp-file
	 (old-shell (sys:getenv "SHELL"))
	 )
        ;; Force dc-run-program to parse command using /bin/sh
        ;; Using /bin/csh courses "Too many )'s." error
        (setf (sys:getenv "SHELL") "/bin/sh")
	(dc-run-program (format nil 
				"~a/bin/dc-~a-checker" 
				*dc-directory* 
				(sys:getenv "PVSARCH"))
			:input stdin-tmp-file 
			:output stdout-tmp-file :if-output-exists :overwrite 
;			:error-output stderr-tmp-file :if-error-output-exists :overwrite
			)
        (setf (sys:getenv "SHELL") old-shell)
	(dc-get-bdd-output-string bdd-out-tmp-file)))


;;; Works on SunOS and allegro 4.2
;;;  (setf (sys:getenv "SHELL") "/bin/sh") does not work on acl4.2
#-(version>= 4 3)
(defun dc-bdd-output (bdd-string)
  (let* ((bdd-out-tmp-file
	  (write-to-temp-file ""))
         ;; this is just to get the name of the bddoutfile
	 (stdin-tmp-file 
	  (write-to-temp-file (format nil "~a \"~a\"" bdd-string bdd-out-tmp-file)))
	 (stdout-tmp-file
	  (write-to-temp-file ""))
         ;; this is just to get the name of the stdout-tmp-file
;	 (stderr-tmp-file
;	  (write-to-temp-file ""))
         ;; this is just to get the name of the stderr-tmp-file
	 )
	(dc-run-program (format nil 
				"~a/bin/dc-~a-checker" 
				*dc-directory* 
				(sys:getenv "PVSARCH"))
			:input stdin-tmp-file 
			:output stdout-tmp-file :if-output-exists :overwrite 
;			:error-output stderr-tmp-file :if-error-output-exists :overwrite
			)
	(dc-get-bdd-output-string bdd-out-tmp-file)))





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions to add a valid assertion to the goal sequent.  
;; NB!! These functions should be used carefully, since they may jeopardize 
;; the soundness of the prover.

;;; Renamed from add-fact during porting to avoid name clash with PVS2
(addrule 'dc-add-fact (fmla) () (dc-add-fact-rule-fun fmla)
	 "(dc-add-fact fmla)"
	 "~%Adding fact")
  
;;; Renamed from add-fact-rule-fun during porting to avoid name clash with PVS2
(defun dc-add-fact-rule-fun (fmla)
  #'(lambda (ps)
      (let* 
	  ((sform (make-instance 's-formula
				 'formula (negate fmla)))
	   (newsequent (lcopy (current-goal ps)
			      's-forms (cons sform
					     (s-forms
					      (current-goal ps))))))
	(values '? (list newsequent)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This is the strategy which is used to apply the decision procedure.

(defdcstep dc-decide (&optional (fnums *))
    (if (find-all-dc-bdd-fmlas '* fnums)
	(then (flatten)
	      (let ((seq (select-seq (s-forms *goal*) (find-all-dc-bdd-fmlas '* fnums)))
		(dummy (initialize-dc-bdd))
		    (enums (enum-lists-from-goal seq))
		    (enumsstr
		     (format nil "[~a]" (mk-rec-string "," ""
						       (mapcar #'(lambda(y) (format nil "[~a]" (mk-rec-string "," ""
													      (mapcar #'(lambda(x)
															  (format nil "\"~a\"" x))
														      y))))
							       enums))))
		    (inputstr (format nil "pvs_valid (\"~a\") ~a" (mk-dc-bdd-string seq) enumsstr))
		    (outputstr (dc-bdd-output inputstr)))
		(if (string-equal "Valid" outputstr)
		    (let ((fact (make-pvs-bdd-string 
				 (find-dc-interval)
				 (select-seq (s-forms *goal*) '*))))
		      (then* (dc-add-fact fact)(do-rewrite)(quant-interval)(ground)))
		  (skip-msg "A counter example exists" ))))
;;      		  (skip-msg outputstr))))
      (skip-msg "The goal sequent does not contain any formulas which conform to the syntax of the DC decision procedure"))
 "
<fnums> designates the formulas on with to run the decision procedure.

Applies the DC decision precedure on the formulas <fnums>."
  "Applying the DC decision procedure"
  misc 
  "Apply DC decision procedure")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Create list of lists of constructors for each enumeration type
;;;  occurring in the expression to be decided.  This must be passed
;;;  to the decision procedure.

;; Given a sequent, return a list of the enumeration constants used in the 
;; sequent.
(defun enum-lists-from-goal (seq) 
   (let ((enumlists (mapcar #'(lambda (x) (create-enum-lists (formula x))) seq)))
      (unique (combine enumlists) #'equal)))


;; Given an expression, create lists of all enumeration literals of enumeration
;; types found in the expression.
(defun create-enum-lists (expr)
  (mapcar #'(lambda(x) (mapcar #'(lambda(y) (make-dc-bdd-variable (make-dc-value-state-equality (caar x)  y))) (cdr x)))
	      (unique (find-enum-types (find-pr-eq-args expr)) #'tc-eq)))

(defun combine (ss)
  (cond ((null ss) nil)
	(t (nconc (car ss) (combine (cdr ss))))))

;; Given an expression, return all the lifted arguments to equalities. 
(defun find-pr-eq-args (expr)
       (cond ((match-pattern? "a_=a_" expr "a_")
	      (let ((arg1 (select-match "a_=b_" expr "a_"))
		    (arg2  (select-match "a_=b_" expr "b_")))
		(if (is-dc-value-state-n? (car arg2))
		    (list (cons arg1 arg2))
		    (list (cons arg2 arg1)))))
     	     ((application? expr)
	      (nconc (find-pr-eq-args (operator expr))
		     (combine (mapcar #'find-pr-eq-args (arguments expr)))))
	     (T nil)))


;; Given a list of expression, return the enumeration types found there.
; An expr having no enum types does not contribute to the result.
(defun find-enum-types (exprs)
  (when exprs
	(let* ((x (car exprs))
	       (et (when (is-dc-enum? (caar x))
			 (cons (car x) 
			       (constructors (adt
					      (range (find-supertype (type (caar x)))))))))
	       (ets (find-enum-types (cdr exprs))))
	  (if et
	      (cons et ets)
	    ets))))

;; A amateur way of making a list of unuque elements (a "set") from a list
;; of elements
(defun unique (list test)
  (unique* list nil test))

(defun unique* (list accumulator test)
  (cond ((null list) accumulator)
	(t (if (assoc (caar list) accumulator :test test)
	       (unique* (cdr list) accumulator test)
	       (unique* (cdr list) (cons (car list) accumulator) test)))))


;; A counter used to create new names for state expressions.
(newcounter *dc-bdd-counter*)

; Given an expression, create a new name of form "DCVARn" for a new number n,
; and push the pair (expr . DCVARn) onto *dc-bdd-assoc-list* and
; (DCVARn . expr) onto *bdd-dc-assoc-list*. 
(defun make-dc-bdd-variable (expr)
  (let ((pair (assoc expr *dc-bdd-assoc-list* :test #'tc-eq)))
  (if pair
      (cdr pair)
    (let ((newname (intern (format nil "DCVAR~a" (funcall *dc-bdd-counter*)))))
      (setq *dc-bdd-assoc-list* (cons (cons expr newname) *dc-bdd-assoc-list*))
      (setq *bdd-dc-assoc-list* (cons (cons newname expr) *bdd-dc-assoc-list*))
      newname))))


; Given expression a=b, make sure that the expressions a=b and b=a are inserted
; in the BDD association lists. Use same method as for make-dc-bdd-variable.
; This is to maintain commutativity of =.
(defun make-dc-bdd-equality-variable (expr)
  (let* ((newexpr (make-dc (format nil "~a=~a" (args2 expr) (args1 expr)) "State[bool]"))
	(pair (or (assoc expr *dc-bdd-assoc-list* :test #'tc-eq)
		  (assoc newexpr *dc-bdd-assoc-list* :test #'tc-eq))))
  (if pair
      (cdr pair)
    (let ((newname (intern (format nil "DCVAR~a" (funcall *dc-bdd-counter*)))))
      (setq *dc-bdd-assoc-list* (cons (cons expr newname) *dc-bdd-assoc-list*))
      (setq *dc-bdd-assoc-list* (cons (cons newexpr newname) *dc-bdd-assoc-list*))
      (setq *bdd-dc-assoc-list* (cons (cons newname expr) *bdd-dc-assoc-list*))
      newname))))


;; Initialize the variables
(defun initialize-dc-bdd ()
    (newcounter *dc-bdd-counter*)
    (setq *dc-bdd-assoc-list* nil)
    (setq *bdd-dc-assoc-list* nil))
