;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;               PROOF CHECKER FOR DURATION CALCULUS IN PVS
;;; File: dc-strategies.lisp
;;; Date: Tue Sep 29 1998
;;; Original version for PVS1 by:
;;;   Jens Ulrik Skakkebaek and Sam Owre
;;; Adapted for PVS2 by:
;;;   Soeren T. Heilmann, sh@it.dtu.dk
;;;   Dept. of Information Technology, Technical University of Denmark
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;        DC PROOF COMMANDS                                             ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'pvs)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DEFINING DC STRATEGIES

;;; Contains all known groups of dcsteps and their UG section headline
(defconstant *dcstep-groups* 
  (list 
   (cons 'misc "Miscellaneous") 
   (cons 'propchop "Basic Properties of Chop")
   (cons 'modal "Modal Proof Commands")
   (cons 'mono "Monotonicity")
   (cons 'quant "Quantifiers")
   (cons 'induct "Induction")
   (cons 'free "Context Free Formulas")
   (cons 'propdur "Duration Properties")))

;;; List associating dcstep name to (group . short)
;;; Build incrementally by defdcstep
(defvar *dcsteps* nil)

;;; defdcstep is an extension of defstep storing in *dcsteps*
;;; information on group membership and a short description string
;;; for each proof command. This information is used in dc-doc.lisp
;;; to derive the User Guide documentation for the commands.

(defmacro defdcstep (name args body doc format group short)
  `(cond ((not (symbolp (quote ,name)))
	  (format T "~%Name ~a must be a symbol." (quote ,name)))
         ((not (symbolp (quote ,group)))
	  (format T "~%Group ~a must be a symbol." (quote ,group)))
         ((not (assoc (quote ,group) *dcstep-groups*))
	  (format T "~%~a is an unknown dcstep group" (quote ,group)))
         ((not (stringp ,short))
	  (format T "~%Short documentation ~a is not a string (in double quotes)."
	    ,short))
         (t 
	  (pushnew (cons (quote ,name) (cons (quote ,group) ,short)) 
		   *dcsteps* 
		   :test #'equal)
	  (defstep ,name ,args ,body ,doc ,format))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; HELPING FUNCTIONS

(defdcstep dc-init () 
  (then* (auto-rewrite-theory "dc_bool") 
         (auto-rewrite-theory "dc_real") 
	 (auto-rewrite-theory "dc_also_rewrite") 
	 ;; Auto-rewrite dc_rules_defs to expand + and * during rule application
	 (auto-rewrite-theory "dc_rules_defs" :defs t)
	 (auto-rewrite! "ex" "fall")
	 (expand "dc_bool.valid")
 	 (assert)
	 (skosimp*))
  "
Initializes the prover for proving DC theorems. Installs
automatic rewriting of DC operators etc.

A customized definition of this proof command may be kept
in the file pvs-strategies in the current context directory.
Such a definition will override this one."
  "Initializing the DC formula"
  misc 
  "Initialize DC goal")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; EXPANDING SHORTHANDS
;;;
;;; Previously, EXPAND automatically invoked automatic rewrite rules,
;;; which are necesary for PC/DC.  However, it is no longer the case,
;;; and therefore is a special version necessary.

(defdcstep dc-expand (name &optional (fnums *))
  (then* 
   (expand name fnums)
   (do-rewrite fnums))
  "
Expands definitions in PC/DC and does rewriting."
  "DC expanding ~a"
  misc 
  "Expand DC definition")

;;;;;; find-dc-context-interval quant-interval delete-dc-fmlas find-dc-pattern
;;;;;; find-dc-fmla delete-dc-seq collect-dc-antec collect-dc-conseq
;;;;;; find-all-dc-pattern make-dc-rec dc-args dc-operators find-pvs-fmla
;;;;;; gen-pc-parse hide-index expand-dc-rule-defs gather-not-context-free
;;;;;; find-context-free-conj
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; BRINGING IN DURATION CALCULUS LEMMAS

(defdcstep dc-lemma (name &optional subst i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(then* (lemma name subst)
	       (do-rewrite -1)
	       (quant-interval i))
      (skip-msg "Could not find DC context interval")))
  "
Adds DC lemma named <name> to the premiss"
  "Introducing lemma ~a"
  misc 
  "Introduce DC lemma")

(defstep equiv (st1 st2)
  (let ((st (format nil "not((~a) = (~a))" st1 st2)))
    (then* (case st)
	   (apply (APPLY-EXTENSIONALITY) (IFF) (GROUND))
	   ))
  "equiv"
  "equiv")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; COMMON FUNCTION USED TO BRING IN AND INSTANTIATE PROOF RULE

(defstep expand-dc-rule-defs (&optional (fnum -1))
  (then* 
   (expand "dc_rules_defs.rule" fnum)
   (expand "dc_rules_defs.seq" fnum)
   (expand "dc_rules_defs.+" fnum)
   (expand "dc_rules_defs.*" fnum))
  "Expands the DC rule definitions"
  "")

(defstep apply-dc-rule (name insts i)
  (then* 
   (branch (lemma name insts)
	   ((then* 
             (do-rewrite -1)
	     (branch (split -1)
		     ((then* ; Prove rule consequence
		       (do-rewrite -1)
		       (quant-interval i)
		       (assert))
		      (then* ; Setup rule premisses
		       (delete-dc-fmlas i)
		       (skolem! 1)
		       (do-rewrite)
                       (ground)
		       (assert)))))
	    (then*  ; For potential TCCs
	      (ground) 
	      (assert))
	    )))
  "
Applies the DC rule <name> with the instantiations <terms>"
  "Applying the DC rule ~a")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; STRATEGIES FOR DURATION CALCULUS PROOF RULES

(defdcstep false-chop (&optional (fnumsl -) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "FALSE;a_" '-)))
	  (if (car nums)
	      (let ((chop (find-dc-fmla  nums))
		    (a (args2 chop))
		    (s (delete-dc-seq nums))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "A" a "Delta" d "Gamma" g)))
		(apply-dc-rule$ "false_chop" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a formula in the antecedent of the form FALSE;A.
<i> is the context interval.

Propagates FALSE to the left of the chopped formula <fnumsl> through 
to the antecedent thus proving the goal."
  "Propagating left false through chop"
  propchop 
  "Propagate left false through chop")



(defdcstep chop-false (&optional (fnumsl -) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "a_;FALSE" '-)))
	  (if (car nums)
	      (let ((chop (find-dc-fmla nums))
		    (a (args1 chop))
		    (s (delete-dc-seq nums))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "A" a "Delta" d "Gamma" g)))
		(apply-dc-rule$ "chop_false" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a formula in the antecedent of the form A;FALSE.
<i> is the context interval.

Propagates FALSE to the right of the chopped formula <fnumsl> through 
to the antecedent thus proving the goal."
  "Propagating right false through chop"
  propchop 
  "Propagate right false through chop")



(defdcstep chop-false-chop (&optional (fnumsl *) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "a_;FALSE;a_" '-)))
	  (if (car nums)
	      (let ((chop (find-dc-fmla nums))
		    (a (args1 (args1 chop)))
		    (b (args2 chop))
		    (s (delete-dc-seq nums))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "A" a "B" b "Delta" d "Gamma" g)))
		(apply-dc-rule$ "chop_false_chop" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "No DC formulas in sequent")))
  "
<fnumsl> designates a formula in the antecedent of the form A;FALSE;B.
<i> is the context interval.

Propagates FALSE in the center of the chopped formula <fnumsl> through 
to the antecedent thus proving the goal."
  "Propagating center false through chop"
  propchop 
  "Propagate center false through chop")



(defdcstep box-right (&optional (fnumsr +) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsr "[](a_)" '+)))
	  (if (car nums)
	      (let ((a (args1 (find-dc-fmla nums)))
		    (s (select-seq (s-forms *goal*) '*))
		    (gs (gather-seq s (find-all-dc-pattern '* "[](a_)" '-) nil))
		    (ds (gather-seq s (find-all-dc-pattern '* "<>(a_)" '+) nil))
		    (g (make-dc-rec "/\\" "TRUE" "IForm"
				    (dc-args (dc-operators (dc-args (formulas gs))))))
		    (d (make-dc-rec "\\/" "FALSE" "IForm"
				    (dc-args (dc-operators (formulas ds)))))
		    (insts (list "A" a "Gamma" g "Delta" d)))
		(apply-dc-rule$ "box_right" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsr> designates a boxed formula in the consequent of the form []A.
<i> is the context interval.

Removes box from formula <fnumsr>. 
Unboxed formulas of the antecedent are removed.
Undiamonded formulas of the consequent are removed.

Use box-fby, box-lto and box-upto to retain formulas of the
antecedent using these operators."
  "Removing consequent box"
  modal 
  "Remove consequent box")



(defdcstep box-left (&optional (fnumsl -) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "[](a_)" '-)))
	  (if (car nums)
	      (let ((a (args1 (find-dc-fmla  nums)))
		    (s (delete-dc-seq nums))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "A" a "Gamma" g "Delta" d)))
		(apply-dc-rule$ "box_left" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a boxed formula in the antecedent of the form []A.
<i> is the context interval.

Removes box from formula <fnumsl>."
  "Removing antecedent box"
  modal 
  "Remove antecedent box")



(defdcstep diamond-left (&optional (fnumsl -) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "<>(a_)" '-)))
	  (if (car nums)
	      (let ((a (args1 (find-dc-fmla nums)))
		    (s (select-seq (s-forms *goal*) '*))
		    (gs (gather-seq s (find-all-dc-pattern '* "[](a_)" '-) nil))
		    (ds (gather-seq s (find-all-dc-pattern '* "<>(a_)" '+) nil))
		    (g (make-dc-rec "/\\" "TRUE" "IForm"
				    (dc-args (dc-operators (dc-args (formulas gs))))))
		    (d (make-dc-rec "\\/" "FALSE" "IForm"
				    (dc-args (dc-operators (formulas ds)))))
		    (insts (list "A" a "Gamma" g "Delta" d)))
		(apply-dc-rule$ "diamond_left" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a diamonded formula in the antecedent of the form <>A.
<i> is the context interval.

Removes diamond from formula <fnumsl>. 
Unboxed formulas of the antecedent are removed.
Undiamonded formulas of the consequent are removed.

Use box-fby, box-lto and box-upto to retain formulas of the
antecedent using these operators."
  "Removing antecedent diamond"
  modal 
  "Remove antecedent diamond")



(defdcstep diamond-right (&optional (fnumsr +) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsr "<>(a_)" '+)))
	  (if (car nums)
	      (let ((a (args1 (find-dc-fmla nums)))
		    (s (delete-dc-seq nums))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "A" a "Gamma" g "Delta" d)))
		(apply-dc-rule$ "diamond_right" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsr> designates a diamonded formula in the consequent of the form <>A.
<i> is the context interval.

Removes diamond from formula <fnumsr>."
  "Removing consequent diamond"
  modal 
  "Remove consequent diamond")



(defdcstep box-fby (&optional (fnums *) (dir LR) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(if (eq dir 'LR)
	    (let ((fby-nums (find-dc-pattern fnums "fby(a_, a_)" '*)))
	      (if (car fby-nums)
		  (let ((fby (find-dc-fmla fby-nums))
			(a (args1 fby))
			(b (args2 fby))
			(insts (list "A" a "B" b)))
		    (rewrite "box_fby" fby-nums insts * dir))
		(skip-msg "Could not find matching DC formula")))
	  (let (
		(fby-nums (find-dc-pattern fnums "[](fby(a_, a_))" '*)))
	    (if (car fby-nums)
		(let ((fby (find-dc-fmla fby-nums))
		      (b_eq (find-pvs-fmla e-nums))
		      (a (args1 (args1 fby)))
		      (b (args1 (args2 fby)))
		      (insts (list "A" a "B" b)))
		  (rewrite "box_fby" fby-nums insts * dir))
	      (skip-msg "Could not find matching DC formula"))))
      (skip-msg "No DC formulas in sequent")))
  "
<fnums> is a formula of the form A-->B or [](A-->B) depending of <dir>.
<dir> is the direction of rewriting.
<i> is the context interval.

Rewrites <fnums> using the equivalence:
  A-->B <=> [](A-->B)

If <dir> is LR boxes <fnums>.
If <dir> is RL unboxes <fnums>."
  "Boxing or unboxing -->"
  modal 
  "Box followed by")



(defdcstep box-lto (&optional (fnums *) (dir LR) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(if (eq dir 'LR)
	    (let (
		  (lto-nums (find-dc-pattern fnums "lto(a_, a_, a_)" '*)))
	      (if (car lto-nums)
		  (let ((lto (find-dc-fmla lto-nums))
			(a (args1 lto))
			(t (args2 lto))
			(b (caddr (arguments lto)))
			(insts (list "A" a "B" b "t" t)))
		    (rewrite "box_lto" lto-nums insts * dir))
		(skip-msg "Could not find matching DC formula")))
	  (let (
		(lto-nums (find-dc-pattern fnums "[](lto(a_, a_, a_))" '*)))
	    (if (car lto-nums)
		(let ((lto (find-dc-fmla lto-nums))
		      (b_eq (find-pvs-fmla e-nums))
		      (a (args1 (args1 lto)))
		      (t (args1 (args2 lto)))
		      (b (args1 (caddr (arguments lto))))
		      (insts (list "A" a "B" b "t" t)))
		  (rewrite "box_lto" lto-nums insts * dir))
	      (skip-msg "Could not find matching DC formula"))))
      (skip-msg "No DC formulas in sequent")))
  "
<fnums> is a formula of the form A-:t:->B or [](A-:t:->B) depending of <dir>.
<dir> is the direction of rewriting.
<i> is the context interval.

Rewrites <fnums> using the equivalence:
  A-:t:->B <=> [](A-:t:->B)

If <dir> is LR boxes <fnums>.
If <dir> is RL unboxes <fnums>."
  "Boxing or unboxing -::->"
  modal 
  "Box leads to")



(defdcstep box-upto (&optional (fnums *) (dir LR) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(if (eq dir 'LR)
	    (let (
		  (lto-nums (find-dc-pattern fnums "upto(a_, a_, a_)" '*)))
	      (if (car lto-nums)
		  (let ((lto (find-dc-fmla lto-nums))
			(a (args1 lto))
			(t (args2 lto))
			(b (caddr (arguments lto)))
			(insts (list "A" a "B" b "t" t)))
		    (rewrite "box_lto" lto-nums insts * dir))
		(skip-msg "Could not find matching DC formula")))
	  (let (
		(lto-nums (find-dc-pattern fnums "[](upto(a_, a_, a_))" '*)))
	    (if (car lto-nums)
		(let ((lto (find-dc-fmla lto-nums))
		      (b_eq (find-pvs-fmla e-nums))
		      (a (args1 (args1 lto)))
		      (t (args1 (args2 lto)))
		      (b (args1 (caddr (arguments lto))))
		      (insts (list "A" a "B" b "t" t)))
		  (rewrite "box_lto" lto-nums insts * dir))
	      (skip-msg "Could not find matching DC formula"))))
      (skip-msg "No DC formulas in sequent")))
  "
<fnums> is a formula of the form A-<=:t:->B or [](A-<=:t:->B) depending of <dir>.
<dir> is the direction of rewriting.
<i> is the context interval.

Rewrites <fnums> using the equivalence:
  A-<=:t:->B <=> [](A-<=:t:->B)

If <dir> is LR boxes <fnums>.
If <dir> is RL unboxes <fnums>."
  "Boxing or unboxing -<=::->"
  modal 
  "Box upto")



(defdcstep mono (&optional (fnumsl -)(fnumsr +) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-))
	      (numsr (find-dc-pattern fnumsr "a_;a_" '+)))
	  (if (and (car numsl) (car numsr))
	      (let ((chopl (find-dc-fmla numsl))
		    (a (args1 chopl))
		    (b (args2 chopl))
		    (chopr (find-dc-fmla numsr))
		    (c (args1 chopr))
		    (d (args2 chopr))
		    (s (delete-dc-seq numsl))
		    (g (collect-dc-antec i s))
		    (s2 (delete-dc-seq numsr))
		    (del (collect-dc-conseq i s2))
		    (insts (list "A" a "B" b "C" c "D" d "Delta" del "Gamma" g)))
		(apply-dc-rule$ "mono" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a chopped formula A;B of the antecedent.
<fnumsr> designates a chopped formula C;D of the consequent.
<i> is the context interval.

Applies monotonicity of chop leaving formulas [](A=>C) and 
[](B=>D) to be proved. The sequent is otherwise preserved."
  "Applying monotonicity"
  mono 
  "Apply monotonicity of chop")



(defdcstep mono! (&optional (fnumsl -) (fnumsr +) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-))
	      (numsr (find-dc-pattern fnumsr "a_;a_" '+)))
	  (if (and (car numsl) (car numsr))
	      (let ((chopl (find-dc-fmla  numsl))
		    (a (args1 chopl))
		    (b (args2 chopl))
		    (chopr (find-dc-fmla numsr))
		    (c (args1 chopr))
		    (d (args2 chopr))
		    (insts (list "A" a "B" b "C" c "D" d)))
		(apply-dc-rule$ "mono_bang" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a chopped formula A;B of the antecedent.
<fnumsr> designates a chopped formula C;D of the consequent.
<i> is the context interval.

Applies monotonicity of chop leaving two subgoals to prove formulas 
A=>C and B=>D. Any other DC formulas of the sequent are removed."
  "Applying monotonicity"
  mono 
  "Apply monotonicity of chop")



(defdcstep mono-chop (&optional (dcfmla "") (fnumsl -) cnum i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-)))
	  (if (car numsl)
	      (then
	       (if cnum (top-assoc cnum numsl) (skip))
	       (let ((chopl (find-dc-fmla numsl))
		     (a (args1 chopl))
		     (b (args2 chopl))
		     (c (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (d (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "Delta" d "Gamma" g)))
		 (apply-dc-rule$ "mono_chop" insts i)))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the left formula A leaving 2 subgoals. 
One introducing F;B into the antecedent, the other proving [](A=>F). 
The DC formulas of the current goal is maintained in both subgoals."
  "Applying left monotonicity of chop by introducing ~a"
  mono 
  "Apply left monotonicity of chop")



(defdcstep mono-chop! (&optional (dcfmla "")(fnumsl -) cnum i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-)))
	  (if (car numsl)
	      (then
	       (if cnum (top-assoc cnum numsl) (skip))
	       (let ((chopl (find-dc-fmla numsl))
		     (a (args1 chopl))
		     (b (args2 chopl))
		     (c (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (d (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "Delta" d "Gamma" g)))
		 (apply-dc-rule$ "mono_chop_bang" insts i)))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the left formula A leaving 2 subgoals. 
One proving A=>F the other introducing F;B into the antecedent. 
The DC formulas of the current goal is maintained only in the second subgoal."
  "Applying left monotonicity of chop and removing box"
  mono 
  "Apply left monotonicity of chop")



(defdcstep chop-mono (&optional (dcfmla "") (fnumsl -) cnum i)
  (let ((i (find-dc-interval)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-)))
	  (if (car numsl)
	      (then
	       (if cnum (top-assoc cnum numsl) (skip))
	       (let ((chopl (find-dc-fmla  numsl))
		     (a (args1 chopl))
		     (b (args2 chopl))
		     (c (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (d (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "Delta" d "Gamma" g)))
		 (apply-dc-rule$ "chop_mono" insts i)))
  	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the right formula B leaving 2 subgoals. 
One introducing A;F into the antecedent, the other proving [](B=>F). 
The DC formulas of the current goal is maintained in both subgoals."
  "Applying right monotonicity of chop by introducing ~a"
  mono 
  "Apply right monotonicity of chop")



(defdcstep chop-mono! (&optional (dcfmla "") (fnumsl -) cnum i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_" '-)))
	  (if (car numsl)
	      (then
	       (if cnum (top-assoc cnum numsl) (skip))
	       (let ((chopl (find-dc-fmla numsl))
		     (a (args1 chopl))
		     (b (args2 chopl))
		     (c (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (d (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "Delta" d "Gamma" g)))
		 (apply-dc-rule$ "chop_mono_bang" insts i)))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the right formula B leaving 2 subgoals. 
One proving B=>F the other introducing A;F into the antecedent. 
The DC formulas of the current goal is maintained only in the second subgoal."
  "Applying right monotonicity of chop and removing box"
  mono 
  "Apply right monotonicity of chop")


(defdcstep chop-mono-chop (&optional (dcfmla "") (fnumsl -) cnuml cnumr i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "(a_;a_);a_" '-)))
	  (if (car numsl)
	      (then
	       (if (and cnuml cnumr) (center-assoc cnuml cnumr numsl) (skip))
	       (let ((chopl (find-dc-fmla numsl))
		     (a (args1 (args1 chopl)))
		     (b (args2 (args1 chopl)))
		     (c (args2 chopl))
		     (d (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (del (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "D" d "Delta" del "Gamma" g)))
		 (apply-dc-rule$ "chop_mono_chop" insts i)))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B;C of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the center interval B leaving 2 subgoals. 
One introducing A;F;C into the antecedent, the other proving [](B=>F). 
The DC formulas of the current goal is maintained in both subgoals.

This proof command was previously called mono-chop-mono."
  "Applying center monotonicity of chop"
  mono 
  "Apply center monotonicity of chop")



(defdcstep chop-mono-chop! (&optional (dcfmla "") (fnumsl -) cnuml cnumr i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "a_;a_;a_" '-)))
	  ;;  (let ((numsl (find-dc-pattern fnumsl "a_^a_^a_" '-)))
	  (if (car numsl)
	      (then
	       (if (and cnuml cnumr) (center-assoc cnuml cnumr numsl) (skip))
	       (let ((chopl (find-dc-fmla  numsl))
		     (c (args2 chopl))
		     (b (args2 (args1 chopl)))
		     (a (args1 (args1 chopl)))
		     (d (gen-pc-parse "IForm" dcfmla))
		     (s (delete-dc-seq numsl))
		     (g (collect-dc-antec i s))
		     (dl (collect-dc-conseq i s))
		     (insts (list "A" a "B" b "C" c "D" d "Delta" dl "Gamma" g)))
		 (apply-dc-rule$ "chop_mono_chop_bang" insts i)))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula F.
<fnumsl> designates a chopped formula A;B;C of the antecedent.
<i> is the context interval.

Applies monotonicity of chop on the center formula B leaving 2 subgoals. 
One proving B=>F the other introducing A;F;C into the antecedent. 
The DC formulas of the current goal is maintained only in the second subgoal.

This proof command was previously called mono-chop-mono!."
  "Applying center monotonicity of chop and removing box"
  mono 
  "Apply center monotonicity of chop")



(defdcstep assoc (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "(a_;a_);a_" '*)))
	(if (car numsl)
	    (let ((chopl (find-dc-fmla  numsl))
		  (a (args1 (args1 chopl)))
		  (b (args2 (args1 chopl)))
		  (c (args2 chopl))
		  (insts (list "A" a "B" b "C" c)))
	      (rewrite "assoc_dc" numsl insts * dir))
      	    (skip-msg "Could not find matching DC formula")))
      (let ((numsl (find-dc-pattern fnums "a_;(a_;a_)" '*)))
	(if (car numsl)
	    (let ((chopl (find-dc-fmla  numsl))
		  (b (args1 (args2 chopl)))
		  (c (args2 (args2 chopl)))
		  (a (args1 chopl))
		  (insts (list "A" a "B" b "C" c)))
	      (rewrite "assoc_dc" numsl insts * dir))
  	    (skip-msg "Could not find matching DC formula"))))
  "
<fnums> designates a formula of the form A;B;C or A;(B;C) depending of <dir>.
<dir> is the direction of rewriting. LR: left to right. RL: right to left.

Applies associativity of chop to rewrite formula <fnums>.
Rewrites A;B;C to A;(B;C) when <dir> is LR.
Rewrites A;(B;C) to A;B;C when <dir> is RL."
  "Applying associativity of chop"
  propchop 
  "Apply associativity of chop")


(defun right-chop-count (fmla)
  (cond 
   ((is-dc-chop? fmla)
    (+ (chop-count (args2 fmla))
       (right-chop-count (args1 fmla))))
   (t 0)))

(defun left-chop-count (fmla)
  (cond 
   ((is-dc-chop? fmla)
    (+ (chop-count (args1 fmla))
       (left-chop-count (args2 fmla))))
   (t 0)))

(defun chop-count (fmla)
  (cond 
   ((is-dc-chop? fmla)
    (+ (chop-count (args1 fmla))
       (chop-count (args2 fmla))
       1))
   (t 0)))

(defstrat iterate (count step)
  (if (eq count 0)
      (skip)
    (let ((count (- count 1)))
      (try step (if (equal (get-goalnum *ps*) 1)
		    (iterate count step)
		  (skip))
	   (skip))))
  "Successively apply STEP along main branch count times or until it does nothing.")



(defdcstep left-assoc (&optional (fnums *) (count))
  (let ((nums (find-dc-pattern fnums "a_;a_" '*)))
    (if (car nums)
	(let ((chop (find-dc-fmla nums))
	      (chopnum (car nums)))	
	  (if count
	      (iterate count (rewrite "assoc" chopnum :dir RL :order OUT))
	    (repeat (rewrite "assoc" chopnum :dir RL :order OUT))))
      (skip-msg "Could not find matching DC formula")))
  "
<fnums> designates a formula.
<count> is a positive integer.

Rewrites chopped formula <fnums> in to left associative form using
the equivalence A;B;C <=> A;(B;C). Rewrites right to left, outside in.
To rewrite left to right use right-assoc.

If <count> is given rewrite <count> number of times else
rewrite repeatedly."
  "Applying left associativity of chop"
  propchop 
  "Left associate chops")



(defdcstep right-assoc (&optional (fnum *) (count))
  (let ((nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chop (find-dc-fmla nums))
	      (chopnum (car nums))
	      (lcc (or lcc (left-chop-count chop))))
	  (if count
	      (iterate count (rewrite "assoc" chopnum :dir LR :order OUT))
	    (repeat (rewrite "assoc" chopnum :dir LR :order OUT))))
      (skip-msg "Could not find matching DC formula")))
  "
<fnums> designates a formula.
<count> is a positive integer.

Rewrites chopped formula <fnums> in to right associative form using
the equivalence A;B;C <=> A;(B;C). Rewrites left to right, outside in.
To rewrite right to left use left-assoc.

If <count> is given rewrite <count> number of times else
rewrite repeatedly."
  "Applying right associativity of chop"
  propchop 
  "Right associate chops")



(defstep chop-left-assoc (count &optional (fnum *))
  (let ((nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chopnum (car nums)))
	  (iterate count (rewrite "chop_assoc" chopnum :dir RL :order OUT)))
      (skip-msg "Could not find matching DC formula")))
  "Rewrites right side of chopped formula into left associcative form."
  "Applying associativity of chop")



(defstep chop-left-assoc-chop (count &optional (fnum *))
  (let ((nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chopnum (car nums)))
	  (iterate count (rewrite "chop_assoc_chop" chopnum :dir RL :order OUT)))
      (skip-msg "Could not find matching DC formula")))
  "Rewrites center of chopped formula into left associcative form."
  "Applying associativity of chop")


(defstep right-assoc-chop (count &optional (fnum *))
  (let ((nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chopnum (car nums)))
	  (iterate count (rewrite "assoc_chop" chopnum :dir LR :order OUT)))
      (skip-msg "Could not find matching DC formula")))
  "Rewrites left side of chopped formula into right associcative form."
  "Applying associativity of chop")


;;; For use in other strategies fnums may be a singular list.
(defdcstep top-assoc (cnum &optional (fnums *))
  (let ((fnum (if (listp fnums) (car fnums) fnums))
	(nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chop (find-dc-fmla nums))
	      (chopnum (car nums))
	      (cc (chop-count chop))
	      (rac (- cc cnum))
	      (clac (- cc (+ cnum 1))))
	  (then*
	   (left-assoc chopnum)
	   (right-assoc chopnum rac)
	   (chop-left-assoc clac chopnum)))
      (skip-msg "Could not find matching DC formula")))
  "
<cnum> designates a chop in <fnums>
<fnums> designates a chopped formula.

Applies associativity of chop to lift the chop designates by <cnum> in the
formula <fnums> to the top. Resulting for example in a formula 
(A;B;C);(D;E;F;G) for <cnum> equal 3.
The left and right side of <cnum> are rewritten into left associative form."
  "Rewriting chop to top"
  propchop 
  "Lift chop to top")


;;; For use in other strategies fnum may be a singular list.
(defdcstep center-assoc (cnuml cnumr &optional (fnum *))
  (let ((fnum (if (listp fnum) (car fnum) fnum))
	(nums (find-dc-pattern fnum "a_;a_" '*)))
    (if (car nums)
	(let ((chop (find-dc-fmla nums))
	      (chopfnum (car nums))
	      (cc (chop-count chop))
	      (rac (- cc cnumr))
	      (clac (- cc (+ cnumr 1)))
	      (rasc (- cnumr (+ cnuml 1)))
	      (clacc (- cnumr (+ cnuml 2))))
	  (then*
	   (left-assoc chopfnum)
	   (right-assoc chopfnum rac)
	   (chop-left-assoc clac chopfnum)
	   (right-assoc-chop rasc chopfnum)
	   (chop-left-assoc-chop clacc chopfnum)))
      (skip-msg "Could not find matching DC formula")))
  "
<cnuml> designates a left chop in <fnums>
<cnumr> designates a right chop in <fnums>
<fnums> designates a chopped formula.

Applies associativity of chop to lift the chops designates by 
<cnuml> and <cnumsr> in the formula <fnums> to the top. 
Resulting for example in a formula 
(A;B);(C;D);(E;F;G) for <cnuml> equal 2 and <cnumr> equal 4.
The left, right and center parts of <fnums> are rewritten into 
left associative form."
  "Rewriting chops to center"
  propchop 
  "Lift chops to top")



(defdcstep exists-intro (state &optional (fnums *) cnum (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_)" '*)))
	(if (car nums)
	    (then
	     (if cnum (top-assoc cnum nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (n (car nums))
		   (b (args2 chop))
		   (a (args1 chop))
		   (p (gen-pc-parse "State[bool]" state))
		   (insts (list "A" a "B" b "p" p)))
	       (then* (rewrite-lemma "exists_intro" insts) (do-rewrite)
					;     	      (then* (rewrite "exists_intro" :fnums n :subst insts) (do-rewrite)
		      (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "exists_intro" :target-fnums fnums :dir dir)
	  (do-rewrite)))
  "
<state> is a boolean state expression.
<fnums> designates a chopped formula A;B or (SOME r: dur(state)=r /\\ A);B 
  depending on <dir>.
<cnum> designates a chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of duration of <state> in the left
formula A producing (SOME r: dur(state) = r /\\ A);B. 
If <cnum> is given then initially this chop is lifted using top-assoc.

If <dir> is RL removes binding of duration of <state> in the left
formula. Producing A;B."
  "Introducing left existential binding of duration measure of ~a"
  quant 
  "Introduce left binding of duration")



(defdcstep intro-exists (state &optional (fnums *) cnum (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_)" '*)))
	(if (car nums)
	    (then
	     (if cnum (top-assoc cnum nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (n (car nums))
		   (b (args2 chop))
		   (a (args1 chop))
		   (p (gen-pc-parse "State" state))
		   (insts (list "B" b "A" a "p" p)))
	       (then* (rewrite-lemma "intro_exists" insts) (do-rewrite) (skolem! n))))
					;	      (then* (rewrite "intro_exists" :fnums n :subst insts) (do-rewrite) (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "intro_exists" :target-fnums fnum :dir dir)
	  (do-rewrite)))
  "
<state> is a boolean state expression.
<fnums> designates a chopped formula A;B or A;(SOME r: dur(state)=r /\\ B) 
  depending on <dir>.
<cnum> designates a chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of duration of <state> in the right
formula B producing A;(SOME r: dur(state) = r /\\ B). 
If <cnum> is given then initially this chop is lifted using top-assoc.

If <dir> is RL removes binding of duration of <state> in the right
formula. Producing A;B."
  "Introducing right existential binding of duration measure of ~a"
  quant 
  "Introduce right binding of duration")



(defdcstep intro-exists-intro (state &optional (fnums *) cnuml cnumr (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_);a_" '*)))
	(if (car nums)
	    (then
	     (if (and cnuml cnumr) (center-assoc cnuml cnumr nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (a (args1 (args1 chop)))
		   (b (args2 (args1 chop)))
		   (c (args2 chop))
		   (n (car nums))
		   (p (gen-pc-parse "State" state))
		   (insts (list "B" b "A" a "C" c "p" p)))
	       (then* (rewrite-lemma "intro_exists_intro"  insts) (do-rewrite) (skolem! n))))
					;  	      (then* (rewrite "intro_exists_intro" :fnums n :subst insts) (do-rewrite) (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "intro_exists_intro" :target-fnums fnums :dir dir)
	  (do-rewrite)))
  "
<state> is a boolean state expression.
<fnums> designates a chopped formula A;B;C or A;(SOME r: dur(state)=r /\\ B);C 
  depending on <dir>.
<cnuml> designates a left chop in the formula <fnums>.
<cnumr> designates a right chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of duration of <state> in the center
formula B producing A;(SOME r: dur(state) = r /\\ B);C. 
If <cnuml> and <cnumr> are given then initially these chops are lifted using coneter-chop.

If <dir> is RL removes binding of duration of <state> in the center
formula. Producing A;B;C."
  "Introducing center existential binding of duration measure of ~a"
  quant 
  "Introduce center binding of duration")



(defdcstep len-intro ( &optional (fnums *) cnum (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_)" '*)))
	(if (car nums)
	    (then
	     (if cnum (top-assoc cnum nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (n (car nums))
		   (b (args2 chop))
		   (a (args1 chop))
		   (insts (list "A" a "B" b)))
	       (then* (rewrite-lemma "len_intro" insts) (do-rewrite)
		      (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "len_intro" :target-fnums fnums :dir dir)
	  (do-rewrite)))
  "
<fnums> designates a chopped formula A;B or (SOME r: len=r /\\ A);B 
  depending on <dir>.
<cnum> designates a chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of length in the left
formula A producing (SOME r: len=r /\\ A);B. 
If <cnum> is given then initially this chop is lifted using top-assoc.

If <dir> is RL removes binding of length in the left formula. Producing A;B."
  "Introducing left existential binding of length"
  quant 
  "Introduce left binding of length")


(defdcstep intro-len (&optional (fnums *) cnum (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_)" '*)))
	(if (car nums)
	    (then
	     (if cnum (top-assoc cnum nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (n (car nums))
		   (b (args2 chop))
		   (a (args1 chop))
		   (insts (list "B" b "A" a)))
	       (then* (rewrite-lemma "intro_len" insts) (do-rewrite) (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "intro_len" :target-fnums fnums :dir dir)
	  (do-rewrite)))
  "
<fnums> designates a chopped formula A;B or A;(SOME r: len=r /\\ B) 
  depending on <dir>.
<cnum> designates a chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of length in the right formula B 
producing A;(SOME r: len=r /\\ B). 
If <cnum> is given then initially this chop is lifted using top-assoc.

If <dir> is RL removes binding of length in the right formula. Producing A;B."
  "Introducing right existential binding of length"
  quant 
  "Introduce right binding of length")



(defdcstep intro-len-intro (&optional (fnums *) cnuml cnumr (dir LR))
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(a_;a_);a_" '*)))
	(if (car nums)
	    (then
	     (if (and cnuml cnumr) (center-assoc cnuml cnumr nums) (skip))
	     (let ((chop (find-dc-fmla  nums))
		   (a (args1 (args1 chop)))
		   (b (args2 (args1 chop)))
		   (c (args2 chop))
		   (n (car nums))
		   (insts (list "B" b "A" a "C" c)))
	       (then* (rewrite-lemma "intro_len_intro"  insts) (do-rewrite) (skolem! n))))
	  (skip-msg "Could not find matching DC formula")))
    (then (rewrite "intro_len_intro" :target-fnums fnums :dir dir)
	  (do-rewrite)))
  "
<fnums> designates a chopped formula A;B;C or A;(SOME r: len=r /\\ B);C 
  depending on <dir>.
<cnuml> designates a left chop in the formula <fnums>.
<cnumr> designates a right chop in the formula <fnums>.
<dir> is the direction of rewriting.

If <dir> is LR introduces binding of length in the center
formula B producing A;(SOME r: len=r /\\ B);C. 
If <cnuml> and <cnumr> are given then initially these chops are 
lifted using center-assoc.

If <dir> is RL removes binding of length in the center formula. 
Producing A;B;C."
  "Introducing center existential binding of length"
  quant 
  "Introduce center binding of length")



(defdcstep exists-chop (&optional (fnums *) (dir LR) type)
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "(ex(a_));a_" '*)))
	(if (car nums)
	    (let ((chop (find-dc-fmla nums))
		  (b (args2 chop))
		  (a (args1 chop))
		  (f (args1 a))
		  (insts (list "B" b "F" f))
		  (lemma (format nil "exists_chop_dc~a"
				 (if type
				     (format nil "[~a]" type)
				   ""))))
	      (then (rewrite lemma :subst insts) (do-rewrite)))
	    (skip-msg "Could not find matching DC formula")))
	      (then (rewrite "exists_chop_dc" :target-fnums fnums :dir dir)
		    (do-rewrite)))
   "
<fnums> is a chopped formula for the form (SOME x:A);B or 
  (SOME x: A;B) depending of <dir>.
<dir> is the direction of rewriting.
<type> is the type of x.

Moves existential quantifier in or out of left hand side of chopped 
formula <fnums>. 
If <dir> is LR the quantifier is moved out producing (SOME x: A;B). 
If <dir> is RL the quantifier is moved in producing (SOME x: A);B.
It may be necessary to supply the type of x in order for PVS to 
instantiate the rewrite rule correctly or otherwise no rewriting will
occur."
  "Moving left existential quantifier"
  quant 
  "Move left $\\exists$ out of chop")



(defdcstep chop-exists (&optional (fnums *) (dir LR) type)
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "a_;(ex(a_))" '*)))
	(if (car nums)
	    (let ((chop (find-dc-fmla  nums))
		  (b (args2 chop))
		  (a (args1 chop))
		  (f (args1 b))
		  (insts (list "A" a "F" f))
		  (lemma (format nil "chop_exists_dc~a"
				 (if type
				     (format nil "[~a]" type)
				   ""))))
	      (then (rewrite lemma :subst insts) (do-rewrite)))
	    (skip-msg "Could not find matching DC formula")))
	      (then (rewrite "chop_exists_dc" :target-fnums fnums :dir dir)
		    (do-rewrite)))
  "
<fnums> is a chopped formula for the form A;(SOME x:B) or 
  (SOME x: A;B) depending of <dir>.
<dir> is the direction of rewriting.
<type> is the type of x.

Moves existential quantifier in or out of right hand side of chopped 
formula <fnums>. 
If <dir> is LR the quantifier is moved out producing (SOME x: A;B). 
If <dir> is RL the quantifier is moved in producing A;(SOME x: B).
It may be necessary to supply the type of x in order for PVS to 
instantiate the rewrite rule correctly or otherwise no rewriting will
occur."
  "Moving right existential quantifier"
  quant 
  "Move right $\\exists$ out of chop")



(defdcstep chop-exists-chop (&optional (fnums *) (dir LR) type)
  (if (eq dir 'LR)
      (let ((nums (find-dc-pattern fnums "a_;(ex(a_));a_" '*)))
	(if (car nums)
	    (let ((chop (find-dc-fmla  nums))
		  (c (args2 chop))
		  (b (args2 (args1 chop)))
		  (a (args1 (args1 chop)))
		  (f (args1 b))
		  (insts (list "A" a "F" f "C" c))
		  (lemma (format nil "chop_exists_chop_dc~a"
				 (if type
				     (format nil "[~a]" type)
				   ""))))
	      (then (rewrite lemma :subst insts) (do-rewrite)))
	    (skip-msg "Could not find matching DC formula")))
	      (then (rewrite "chop_exists_chop_dc" :target-fnums fnums :dir dir)
		    (do-rewrite)))
  "
<fnums> is a chopped formula for the form A;(SOME x:B);C or 
  (SOME x: A;B;C) depending of <dir>.
<dir> is the direction of rewriting.
<type> is the type of x.

Moves existential quantifier in or out of center formula of chopped 
formula <fnums>. 
If <dir> is LR the quantifier is moved out producing (SOME x: A;B;C). 
If <dir> is RL the quantifier is moved in producing A;(SOME x: B);C.
It may be necessary to supply the type of x in order for PVS to 
instantiate the rewrite rule correctly or otherwise no rewriting will
occur."
  "Moving center existential quantifier"
  quant 
  "Move center $\\exists$ out of chop")



(defdcstep axiom4 (&optional (fnums *) (dir LR) i)
  (if (eq dir 'LR)
      (let ((i (find-dc-context-interval i)))
	(if i
	    (let ((nums (find-dc-pattern fnums "(dur(a_)=n(a_));(dur(a_)=n(a_))" '*)))
	      (if (car nums)
		  (let ((chop (find-dc-fmla nums))
			(b (args2 chop))
			(a (args1 chop))
			(p (select-match "(dur(b_)=n(a_));(dur(a_)=n(a_))" chop "b_"))
			(r (select-match "(dur(a_)=n(b_));(dur(a_)=n(a_))" chop "b_"))
			(s (select-match "(dur(a_)=n(a_));(dur(a_)=n(b_))" chop "b_"))
			(insts (list "p" p "r" r "s" s "i" i)))
		    (then (rewrite-lemma "axiom4_dc" insts :dir dir) 
			  (do-rewrite)))
		(skip-msg "Could not find matching DC formula")))
	  (skip-msg "Could not find DC context interval")))
    (then (rewrite "axiom4_dc" :target-fnums fnums :dir dir) 
	  (do-rewrite)))
  "
<fnums> designates a formula of the form (dur(p)=n(r));(dur(p)=n(s)) or
  dur(p)=n(r+s) depending of <dir>.
<dir> is the direction of rewriting.
<i> is the context interval.

Rewrites the formula <fnums> using axiom 4 of DC: 
  (dur(p)=n(r));(dur(p)=n(s)) <=> dur(p)=n(r+s)"
  "Rewriting using Axiom 4"
  propdur 
  "Apply axiom 4")



(defdcstep bool-induct-chop (R state &optional i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((s (s-forms *goal*))
	      (g (collect-dc-antec i s))
	      (d (collect-dc-conseq i s))
	      (r (make-dc R "[IForm->IForm]"))
	      (sub (format nil " subst(~a)" r))
	      (insts (list "R" r "s" state "Delta" d "Gamma" g)))
	  (spread (case sub)
		  ((then* (lemma "bool_induct_chop")
			  (expand "rule" -1)
			  (expand "seq" -1)
			  (expand "+" -1)
			  (expand "*" -1)
			  (quant? * insts)
			  (do-rewrite -1)
			  (split)
			  (quant-interval i)
			  (assert)
			  (skosimp*)
			  (hide-index))
		   (then* 
		    (expand "subst")
		    (expand "equiv")
		    (skosimp*)
		    (inst?)
		    (assert)
		    (replace -1)))))
      (skip-msg "Could not find DC context interval")))
  "
<R> is a DC formula schema of type [IForm -> IForm].
<state> is a boolean state expression.
<i> is the context interval.

Uses left DC induction of boolean states to produce three subgoal.
One introducing []R(TRUE) in to the antecedent. 
The other to prove []R(point). The third to prove 
[]([]H(X) => R(X \\/ 'state';X \\/ 'NOT state';X)) for all DC formulas X."
  "Using left DC induction"
  induct 
  "Left induct on bool state")



(defdcstep bool-chop-induct (R state &optional i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((s (s-forms *goal*))
	      (g (collect-dc-antec i s))
	      (d (collect-dc-conseq i s))
	      (r (make-dc R "[IForm->IForm]"))
	      (sub (format nil " subst(~a)" r))
	      (insts (list "R" r "s" state "Delta" d "Gamma" g)))
	  (spread (case sub)
		  ((then* (lemma "bool_chop_induct")
			  (expand "rule" -1)
			  (expand "seq" -1)
			  (expand "+" -1)
			  (expand "*" -1)
			  (quant? * insts)
			  (do-rewrite -1)
			  (split)
			  (quant-interval i)
			  (assert)
			  (skosimp*)
			  (hide-index))
		   (then* 
		    (expand "subst")
		    (expand "equiv")
		    (skosimp*)
		    (inst?)
		    (assert)
		    (replace -1)))))
      (skip-msg "Could not find DC context interval")))
  "
<R> is a DC formula schema of type [IForm -> IForm].
<state> is a boolean state expression.
<i> is the context interval.

Uses right DC induction of boolean states to produce three subgoal.
One introducing []R(TRUE) in to the antecedent. 
The other to prove []R(point). The third to prove 
[]([]H(X) => R(X \\/ X;'state' \\/ X;'NOT state')) for all DC formulas X."
  "Using right DC induction"
  induct 
  "Right induct on bool state")



(defdcstep induct-chop (R state &optional i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((s (s-forms *goal*))
	      (g (collect-dc-antec i s))
	      (d (collect-dc-conseq i s))
	      (r (make-dc R "[IForm->IForm]"))
	      (sub (format nil " subst(~a)" r))
	      (insts (list "R" r "s" state "Delta" d "Gamma" g)))
	  (spread (case sub)
		  ((branch (lemma "induct_chop" insts)
			   ((then* 
			     (delete -2) ; Remove subst(r) formula
			     (expand-dc-rule-defs)
			     (branch (split -1)
				     ((then* ; Prove rule consequence
					     (do-rewrite -1)
					     (quant-interval i)
					     (assert))
				      (then* ; Setup rule premisses
					     (delete-dc-fmlas i)
					     (skolem! 1) ; skolem X
					     (skolem! 1) ; skolem i
					     (do-rewrite)
					     (ground)
					     (assert)))))
			    (then*  ; For potential TCCs
				    (ground) 
				    (assert))
			   ))
		  (then* 
		   (expand "dc_induct_defs.subst")
		   (expand "dc_induct_defs.equiv")
		   (skosimp*)
		   (inst?)
		   (assert)
		   (replace -1)))))
      (skip-msg "Could not find DC context interval")))
  "
<R> is a DC formula schema of type [IForm -> IForm].
<state> is a state expression.
<i> is the context interval.

Uses left DC induction on states to produce three subgoal.
One introducing []R(TRUE) in to the antecedent. 
The other to prove []R(point). The third to prove 
[]([]H(X) => R(X \\/ (SOME x: 'state=m(x)';X))) for all DC formulas X."
  "Using left DC induction"
  induct 
  "Left induct")



(defdcstep chop-induct (R state &optional i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((s (s-forms *goal*))
	      (g (collect-dc-antec i s))
	      (d (collect-dc-conseq i s))
	      (r (make-dc R "[IForm->IForm]"))
	      (sub (format nil " subst(~a)" r))
	      (insts (list "R" r "s" state "Delta" d "Gamma" g)))
	  (spread (case sub)
		  ((branch (lemma "chop_induct" insts)
			   ((then* 
			     (delete -2) ; Remove subst(r) formula
			     (expand-dc-rule-defs)
			     (branch (split -1)
				     ((then* ; Prove rule consequence
					     (do-rewrite -1)
					     (quant-interval i)
					     (assert))
				      (then* ; Setup rule premisses
					     (delete-dc-fmlas i)
					     (skolem! 1) ; skolem X
					     (skolem! 1) ; skolem i
					     (do-rewrite)
					     (ground)
					     (assert)))))
			    (then*  ; For potential TCCs
				    (ground) 
				    (assert))
			   ))
		  (then* 
		   (expand "dc_induct_defs.subst")
		   (expand "dc_induct_defs.equiv")
		   (skosimp*)
		   (inst?)
		   (assert)
		   (replace -1)))))
      (skip-msg "Could not find DC context interval")))
  "
<R> is a DC formula schema of type [IForm -> IForm].
<state> is a state expression.
<i> is the context interval.

Uses right DC induction on states to produce three subgoal.
One introducing []R(TRUE) in to the antecedent. 
The other to prove []R(point). The third to prove 
[]([]H(X) => R(X \\/ (SOME x: X;'state=m(x)'))) for all DC formulas X."
  "Using right DC induction"
  induct 
  "Right induct")



(defdcstep disj-chop (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "(a_\\/a_);a_" '*)))
	(if (car numsl)
	    (let ((chopl (find-dc-fmla numsl))
		  (a (args1 (args1 chopl)))
		  (b (args2 (args1 chopl)))
		  (c (args2 chopl))
		  (insts (list "A" a "B" b "C" c)))
	      (rewrite "disj_chop_dc" :subst insts :dir dir))
	    (skip-msg "Could not find matching DC formula")))
      (rewrite "disj_chop_dc" :target-fnums fnums :dir dir))
  "
<fnums> designates a formula of the form (A\\/B);C or A;C \\/ B;C 
  depending of <dir>.
<dir> is the direction of rewriting.

Distributes chop over left disjunction in <fnums> using the equivalence:
  (A\\/B);C <=> A;C \\/ B;C.
<dir> gives the direction of rewriting."
  "Distributing chop over left disjunction"
  propchop 
  "Distribute chop over left disjuction")



(defdcstep chop-disj (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "a_;(a_\\/a_)" '*)))
	(if (car numsl)
	    (let ((chopl (find-dc-fmla  numsl))
		  (b (args1 (args2 chopl)))
		  (c (args2 (args2 chopl)))
		  (a (args1 chopl))
		  (insts (list "A" a "B" b "C" c)))
	      (rewrite "chop_disj_dc" :subst insts :dir dir))
	  (skip-msg "Could not find matching DC formula")))
    (rewrite "chop_disj_dc" :target-fnums fnum :dir dir))
  "
<fnums> designates a formula of the form A;(B\\/C) or A;B \\/ A;C 
  depending of <dir>.
<dir> is the direction of rewriting.

Distributes chop over right disjunction in <fnums> using the equivalence:
   A;(B\\/C) <=> A;B \\/ A;C
<dir> gives the direction of rewriting."
  "Distributing chop over right disjunction"
  propchop 
  "Distribute chop over right disjuction")

(defdcstep chop-disj-chop (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "a_;(a_\\/a_);a_" '*)))
	(if (car numsl)
	    (let ((chopl (find-dc-fmla  numsl))
		  (b (args1 (args2 (args1 chopl))))
		  (c (args2 (args2 (args1 chopl))))
		  (d (args1 chopl))
		  (a (args1 (args1 chopl)))
		  (insts (list "A" a "B" b "C" c "D" d)))
	      (rewrite "chop_disj_chop_dc" :subst insts :dir dir))
	  (skip-msg "Could not find matching DC formula")))
    (rewrite "chop_disj_chop_dc" :target-fnums fnum :dir dir))
  "
<fnums> designates a formula of the form A;(B\\/C);D or A;B;D \\/ A;C;D 
  depending of <dir>.
<dir> is the direction of rewriting.

Distributes chop over center disjunction in <fnums> using the equivalence:
   A;(B\\/C);D <=> A;B;D \\/ A;C;D
<dir> gives the direction of rewriting."
  "Distributing chop over center disjunction"
  propchop 
  "Distribute chop over center disjuction")



(defdcstep chop-and (&optional (fnumsl *) (fnumsr *) (dir RL) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(if (eq dir 'RL)
	    (let (
		  (numsl (find-dc-pattern fnumsl "a_;(len=n(a_)/\\a_)" '*))
		  (numsr (find-dc-pattern fnumsr "a_;(len=n(a_)/\\a_)" '*)))
	      (if (and (car numsl) (car numsr))
		  (let ((chopl (find-dc-fmla numsl))
			(a (args1 chopl))
			(b (args2 (args2 chopl)))
			(r (args1 (args2 (args1 (args2 chopl)))))
			(chopr (find-dc-fmla numsr))
			(c (args1 chopr))
			(d (args2 (args2 chopr)))
			(insts (list "A" a "B" b "C" c "D" d "r" r)))
		    (then* (lemma "chop_and_dc") (inst? * insts) (do-rewrite) (ground)))
		(skip-msg "Could not find matching DC formula")))
	  (skip-msg "Direction not implemented yet"))
      (skip-msg "No DC formulas in sequent")))
  "
<fnumsl> designates a formula of the form A;(len=r /\\ B)
<fnumsl> designates a formula of the form C;(len=r /\\ D)
<dir> is the direction of rewriting.
<i> is the context interval.

Introduces the equivalence:
  A;(len=r /\\ B) /\\ C;(len=r /\\ D) <=> (A /\\ C);(len=r /\\ B /\\ D)
into the antecedent and does propositional simplification using ground.

Direction LR is not implemented."
  "Applying chop-and"
  propchop 
  "Distribute conjunction over chop")



(defdcstep and-chop (&optional (fnumsl *) (fnumsr *) (dir RL) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(if (eq dir 'RL)
	    (let (
		  (numsl (find-dc-pattern fnumsl "(len=n(a_)/\\a_);a_" '*))
		  (numsr (find-dc-pattern fnumsr "(len=n(a_)/\\a_);a_" '*)))
	      (if (and (car numsl) (car numsr))
		  (let ((chopl (find-dc-fmla numsl))
			(b (args2 chopl))
			(a (args2 (args1 chopl)))
			(r (args1 (args2 (args1 (args1 chopl)))))
			(chopr (find-dc-fmla numsr))
			(d (args2 chopr))
			(c (args2 (args1 chopr)))
			(insts (list "A" a "B" b "C" c "D" d "r" r)))
		    (then* (lemma "and_chop_dc") (inst? * insts) (do-rewrite) (ground)))
		(skip-msg "Could not find matching DC formula")))
	  (skip-msg "Direction not implemented yet"))
      (skip-msg "No DC formulas in sequent")))
  "
<fnumsl> designates a formula of the form (len=r /\\ A);B
<fnumsl> designates a formula of the form (len=r /\\ C);D
<dir> is the direction of rewriting.
<i> is the context interval.

Introduces the equivalence:
  (len=r /\\ A);B /\\ (len=r /\\ C);D <=> (len=r /\\ A /\\ C);(B /\\ D)
into the antecedent and does propositional simplification using ground.

Direction LR is not implemented."
  "Applying and-chop"
  propchop 
  "Distribute conjunction over chop")



(defdcstep free-chop (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "(a_;a_)" '*))
	    (n (car numsl)))
	(if n
	    (then* (auto-rewrite-theory "dc_cf")
		   (do-rewrite n)
		   (stop-rewrite-theory "dc_cf")
		   (let ((chopl (find-dc-fmla numsl))
			 (f (find-context-free-conj (args1 chopl))))
		     (if f
			 (let ((f1 (args1 f))
			       (a (gather-not-context-free (args1 chopl) f))
			       (b (args2 chopl))
			       (instst (list "B" b "f" f1))
			       (insts (list "A" a "B" b "f" f1)))
			   (then* (if (is-dc-true? a)
				      (rewrite "free_chop_true_dc" :subst instst :dir dir)
				      (rewrite "free_chop_dc" :subst insts :dir dir))
				  (auto-rewrite-theory "dc_cf_reverse")
				  (do-rewrite)
				  (stop-rewrite-theory "dc_cf_reverse")))
			 (skip-msg "No left context free expression found"))))
	    (skip-msg "Could not find matching DC formula")))
      (then* (rewrite "free_chop_dc" :dir dir)
	     (auto-rewrite-theory "dc_cf_reverse")
	     (do-rewrite)
	     (stop-rewrite-theory "dc_cf_reverse")))
  "
<fnums> is a formula of the form CF /\\ A;B or CF /\\ (A;B) depending of <dir>.
<dir> is the direction of rewriting.

If CF is a context free formula propagates CF in or out of chopped formula
using rewriting on the equivalence:
  CF /\\ A;B <=> CF /\\ (A;B)
<dir> is the direction of rewriting."
      "Propagating left context free expression through chop"
  free 
  "Propagate left context free expression through chop")



(defdcstep chop-free (&optional (fnums *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnums "(a_;a_)" '*))
	    (n (car numsl)))
	(if n
	    (then* (auto-rewrite-theory "dc_cf")
		   (do-rewrite n)
		   (stop-rewrite-theory "dc_cf")
		   (let ((chopl (find-dc-fmla  numsl))
			 (f (find-context-free-conj (args2 chopl))))
		     (if f
			 (let ((f1 (args1 f))
			       (b (gather-not-context-free (args2 chopl) f))
			       (a (args1 chopl))
			       (instst (list "A" a "f" f1))
			       (insts (list "A" a "B" b "f" f1)))
			   (then* (if (is-dc-true? b)
				      (rewrite "chop_free_true_dc" :subst instst :dir dir)
				      (rewrite "chop_free_dc" :subst insts :dir dir))
				  (auto-rewrite-theory "dc_cf_reverse")
				  (do-rewrite)
				  (stop-rewrite-theory "dc_cf_reverse")))
			 (skip-msg "No right context free expression found"))))
	    (skip-msg "Could not find matching DC formula")))
      (then* (rewrite "chop_free_dc" :dir dir)
	     (auto-rewrite-theory "dc_cf_reverse")
	     (do-rewrite)
	     (stop-rewrite-theory "dc_cf_reverse")))
  "
<fnums> is a formula of the form A;B /\\ CF or CF /\\ (A;B) depending of <dir>.
<dir> is the direction of rewriting.

If CF is a context free formula propagates CF in or out of chopped formula
using rewriting on the equivalence:
  A;B /\\ CF <=> CF /\\ (A;B)
<dir> is the direction of rewriting."
  "Propagating right context free expression through chop"
  free 
  "Propagate right context free expression through chop")



(defdcstep chop-free-chop (&optional (fnumsl *) (dir LR))
  (if (eq dir 'LR)
      (let ((numsl (find-dc-pattern fnumsl "(a_;a_;a_)" '*))
	    (n (car numsl)))
	(if n
	    (then* (auto-rewrite-theory "dc_cf")
		   (do-rewrite n)
		   (stop-rewrite-theory "dc_cf")
		   (let ((chopl (find-dc-fmla  numsl))
			 (f (find-context-free-conj (args2 (args1 chopl)))))
		     (if f
			 (let ((f1 (args1 f))
			       (a (args1 (args1 chopl)))
			       (b (gather-not-context-free (args2 (args1 chopl)) f))
			       (c (args2 chopl))
			       (instst (list "A" a "C" c "f" f1))
			       (insts (list "A" a "B" b "C" c "f" f1)))
			   (then* (if (is-dc-true? b)
				      (rewrite "chop_free_chop_true_dc" :subst instst :dir dir)
				      (rewrite "chop_free_chop_dc" :subst insts :dir dir))
				  (auto-rewrite-theory "dc_cf_reverse")
				  (do-rewrite)
				  (stop-rewrite-theory "dc_cf_reverse")))
			 (skip-msg "No left context free expression found"))))
	    (skip-msg "Could not find matching DC formula")))
      (then* (rewrite "chop_free_chop_dc" :dir dir)
	     (auto-rewrite-theory "dc_cf_reverse")
	     (do-rewrite)
	     (stop-rewrite-theory "dc_cf_reverse")))
   "
<fnums> is a formula of the form A;(B /\\ CF);C or CF /\\ (A;B;C) depending of <dir>.
<dir> is the direction of rewriting.

If CF is a context free formula propagates CF in or out of chopped formula
using rewriting on the equivalence:
  A;(B /\\ CF);C <=> CF /\\ (A;B;C)
<dir> is the direction of rewriting."
      "Propagating center context free expression through chop"
  free 
  "Propagate center context free expression through chop")



(defdcstep dc-case (dcfmla &optional i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((istring (format nil "~a" i))
	      (fmla (format nil "(~a)(~a)" dcfmla i)))
	  (then* (case fmla) (do-rewrite)))
      (skip-msg "Could not find DC context interval")))
  "
<dcfmla> is a DC formula.
<i> is the context interval.

Does a case split on the DC formula <dcfmla> producing two subgoals.
One with <dcfmla> added to the antecedent and one with <dcfmla> 
added to the consequent. "
  "~%DC case splitting on ~%~a"
  misc 
  "Case split on DC formula")
	 


(defdcstep dur-preserve (&optional (fnumsl -) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((nums (find-dc-pattern fnumsl "(a_;a_)" '-)))
	  (if (car nums)
	      (let ((chop (find-dc-fmla nums))
		    (states (get-states chop))
		    (p (args1 (car states)))
		    (q (args1 (cadr states)))
		    (g (make-lambda-from-expr-iform (args1 chop)))
		    (g1 (make-lambda-from-expr-bool (args1 chop)))
		    (s (delete-dc-seq nums))
		    (gam (collect-dc-antec i s))
		    (del (collect-dc-conseq i s))
		    (insts (list "s1" p "s2" q "p" g
				 "Delta" del "Gamma" gam)))
		(branch (lemma "dc_preserve.dur_preserve" insts) 
			((then* (expand "rule" -1)
				(expand "seq" -1)
				(expand "+" -1)
				(expand "*" -1)
				(do-rewrite -1)
				(split)
				(quant-interval i)
				(assert)
				(skosimp*)
				(hide-index))
			 (then* (expand "additive") ; Prove TCCS
				(expand "lifted")
				(instantiate 1 g1)
				(skosimp*)
				(assert)
				(skosimp*)
				(assert)))))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "Could not find DC context interval")))
  "
<fnumsl> designates a formula in the antecedent of the form 
  (dur(p)*dur(q));(dur(p)*dur(q)).
<i> is the context interval.

Applies duration preservation of predicate of durations over chop.
If * is an additive predicate (like =, >, >= etc.) the formula 
dur(p)*dur(q) in the antecedent. <fnumsl> is removed."
  "Using duration preservation"
  propdur 
  "Apply duration preservation over chop")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; STRATEGIES FOR lifting, EXPERIMENTAL

(defstep apply-dc-s-rule (name insts i)
  (then* 
   (branch (lemma name insts)
	   ((then* 
;	     (expand "rule" -1)
;	     (expand "seq" -1)
;	     (expand "sseq" -1)
;	     (expand "+" -1) ; also expands arithmetic + and *, bad thing
;	     (expand "*" -1) ; instead auto-rewrite dc_rules_defs
             (do-rewrite -1)
	     (spread (split -1)
		     ((then* ; Prove rule consequence
		       (do-rewrite -1)
		       (quant-interval)
		       (assert))
		      (then* ; Setup DC rule premiss
		       (delete-dc-fmlas i)
		       (skolem! 1)
		       (do-rewrite)
                       (ground)
		       (assert))
		      (then* ; Setup Math rule premiss
		       (delete-dc-fmlas i)
		       (skolem! 1)
		       (do-rewrite)
		       (ground)
		       (assert)))))
	    (then*  % For potential TCCs
	      (ground) 
	      (assert))
	    )))
  "
Applies the DC rule <name> with the instantiations <terms>"
  "Applying the DC s-rule ~a")
		
(defdcstep assertion (statep stateq i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((p (gen-pc-parse "State" statep))
	      (q (gen-pc-parse "State" stateq))
	      (s (s-forms *goal*))
	      (g (collect-dc-antec i s))
	      (d (collect-dc-conseq i s))
	      (insts (list "Delta" d "Gamma" g "p" p "q" q)))
	  (apply-dc-s-rule$ "assertion" insts i))
      (skip-msg "Could not find DC context interval")))
  " 
<statep> is a boolean state expression.
<stateq> is a boolean state expression.
<i> is the context interval.

Uses assertion on <statep> and <stateq>. Produces two subgoals.
One with dur(statep)=dur(stateq) added to the antecedent.
The other proving (statep<=>stateq)(t) for all times t."
  "~%Using assertion on ~a and ~a."
  propdur 
  "Assert equality on durations")

(defdcstep p-monotone (&optional (fnumsl -) (fnumsr +) i)
  (let ((i (find-dc-context-interval i)))
    (if i
	(let ((numsl (find-dc-pattern fnumsl "pr(a_)" '-))
	      (numsr (find-dc-pattern fnumsr "pr(a_)" '+)))
	  (if (and (car numsl) (car numsr))
	      (let ((p (args1 (find-dc-fmla numsl)))
		    (q (args1 (find-dc-fmla numsr)))
		    (s (select-seq (s-forms *goal*) '*))
		    (g (collect-dc-antec i s))
		    (d (collect-dc-conseq i s))
		    (insts (list "Delta" d "Gamma" g "p" p "q" q)))
		(apply-dc-s-rule$ "p_monotone" insts i))
	    (skip-msg "Could not find matching DC formula")))
      (skip-msg "No DC formulas in sequent")))
  "
<fnumsl> designates a formula in the antecedent of the form 'p'.
<fnumsl> designates a formula in the consequent of the form 'q'.
<i> is the context interval.

Applies monotonicity of state predicates to derive 'p'=>'q' from
proving (p=>q)(t) for all times t.
All DC formulas of the sequent are removed."
  "Using p-monotone"
  propdur 
  "Assert state predicate implication")

(defstep break ()
  (let ((b (break)))
    (skip))
  "
Causes a break to LISP. Use :cont or (restore) to continue."
  "Breaking to LISP. Use (restore) to continue")








