;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;               PROOF CHECKER FOR DURATION CALCULUS IN PVS
;;; File: dc-utils.lisp
;;; Date: Thu Oct 01 1998
;;; Original version for PVS1 by:
;;;   Jens Ulrik Skakkebaek and Sam Owre
;;; Adapted for PVS2 by:
;;;   Soeren T. Heilmann, sh@it.dtu.dk
;;;   Dept. of Information Technology, Technical University of Denmark
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'pvs)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Initial definitions

(defvar *Interval-type* nil)

(defun Interval-type ()
  (or *Interval-type* 
      (setq *Interval-type* (typecheck (pc-parse "Interval" 'type-expr)))))

(defvar *IForm-type* nil)

(defun IForm-type ()
  (or *IForm-type* 
      (setq *IForm-type* (typecheck (pc-parse "IForm" 'type-expr)))))

(defvar *IReal-type* nil)

(defun IReal-type ()
  (or *IReal-type* 
      (setq *IReal-type* (typecheck (pc-parse "IReal" 'type-expr)))))

(defvar *dc-real-type* nil)

(defun dc-real-type ()
  (or *dc-real-type* 
      (setq *dc-real-type* (typecheck (pc-parse "real" 'type-expr)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  PATTERN MATCHING FUNCTIONS

(newcounter *pattern-counter*)

(defun match-pattern? (patternstr expr symbol)
  (let* ((pattern (pc-parse  patternstr 'expr)))
    (match-pattern?* pattern expr (pc-parse symbol 'expr))))


(defun match-pattern?* (pattern expr symbol)
  (cond ((and (name-expr? pattern) (eq (id pattern) (id symbol)))
	 expr)
	((and (name-expr? pattern) (name-expr? expr))
	 (eq (id pattern) (id expr)))
	((and (number-expr? pattern) (number-expr? expr))
	 (eq (id pattern) (id expr)))
	((and (application? pattern) (application? expr))
	 (and (match-pattern?* (operator pattern) (operator expr)
			       symbol)
	      (pattern-and (arguments pattern) (arguments expr)
			   symbol)))
	((and (tuple-expr? pattern) (tuple-expr? expr))
	 (pattern-and (exprs pattern) (exprs expr) symbol))
	((and (binding-expr? pattern) (binding-expr? expr))
	 (break "conditional in match-pattern?* not exhaustive"))
	((and (record-expr? pattern) (record-expr? expr))
	 (break "conditional in match-pattern?* not exhaustive"))
	((and (coercion? pattern) (coercion? expr))
	 (break "conditional in match-pattern?* not exhaustive"))
;;;	     ((and (intype? pattern) (intype? expr))
;;;	           (break "conditional in match-pattern?* not exhaustive"))
	((and (cases-expr? pattern) (cases-expr? expr))
	 (break "conditional in match-pattern?* not exhaustive"))
	((and (update-expr? pattern) (update-expr? expr))
	 (break "conditional in match-pattern?* not exhaustive"))
	(T nil)))

;;;   binding-expr	- bindings, type-value, expression
;;;   record-expr	- assignments
;;;   coercion		- expression, declared-type
;;;   intype		- expression, declared-type, type-value
;;;   cases-expr	- expression, selections, else-part
;;;   update-expr	- expression, assignments


(defun pattern-and (patlist exprlist symbol)
  (cond ((car patlist)
	 (and (match-pattern?* (car patlist) (car exprlist) symbol)
	      (pattern-and  (cdr patlist) (cdr exprlist) symbol)))
	(T T)))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  SELECTION FUNCTIONS

(defun select-match (selectstr expr symbol)
  (let* ((selstr (substitute *dc-chop-char* #\; selectstr))
	 (select (pc-parse  selstr 'expr)))
    (select-match* select expr (pc-parse symbol 'expr))))



(defun select-match* (select expr symbol)
  (cond ((and (name-expr? select) (eq (id select) (id symbol)))
	 (list expr))
;;;	     ((and (name-expr? select) (name-expr? expr)
;;;	           (eq (id select) (id expr)))
;;;	      (list expr))
;;;	     ((and (number-expr? select) (number-expr? expr)
;;;	           (eq (id select) (id expr)))
;;;	      (list expr))
	((and (application? select) (application? expr))
	 (nconc (select-match* (operator select) (operator expr)
			       symbol)
		(select-or (arguments select) (arguments expr)
			   symbol)))
	((or (tuple-expr? select) (tuple-expr? expr))
	 (select-or (exprs select) (exprs expr) symbol))
	((or (binding-expr? select) (binding-expr? expr))
	 (break "conditional in select-match* not exhaustive"))
	((or (record-expr? select) (record-expr? expr))
	 (break "conditional in select-match* not exhaustive"))
	((or (coercion? select) (coercion? expr))
	 (break "conditional in select-match* not exhaustive"))
;;;	     ((or (intype? select) (intype? expr))
;;;	           (break "conditional in select-match* not exhaustive"))
	((or (cases-expr? select) (cases-expr? expr))
	 (break "conditional in select-match* not exhaustive"))
	((or (update-expr? select) (update-expr? expr))
	 (break "conditional in select-match* not exhaustive"))
	(T nil)))

;;;   binding-expr	- bindings, type-value, expression
;;;   record-expr	- assignments
;;;   coercion		- expression, declared-type
;;;   intype		- expression, declared-type, type-value
;;;   cases-expr	- expression, selections, else-part
;;;   update-expr	- expression, assignments


(defun select-or (patlist exprlist symbol)
  (cond ((car patlist)
	 (nconc (select-match* (car patlist) (car exprlist) symbol)
		(select-or  (cdr patlist) (cdr exprlist) symbol)))
	(T nil)))


(defun defined-in? (op theory)
  (eq (id (module-instance (resolution op))) theory))

(defun is-type? (fmla typestr)
  (tc-eq (type fmla) (typecheck (pc-parse typestr 'type-expr))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  

(defstep quant-interval (&optional i subst)
  (if i
      (let ((substs (append `("i" ,i) subst)))
	(quant? * substs))
    (let ((indices (loop for sf in (s-forms *goal*)
			 when (and (is-top-dc-fmla? (formula sf))
				   (dc-index (formula sf)))
			 collect (dc-index (formula sf))))
	  (max-index (when indices (apply #'max (mapcar #'symbol-index indices))))
					;	(max-index 1)
	  (string (format nil "i!~a" max-index))
	  (substs (append `("i" ,string) subst)))
      (if (>= max-index 1)
	  (quant? * substs)
	(quant? * subst))))
  " " " ")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  HIDE INDEX

(defstep hide-index ()
  (let ((indices (loop for sf in (s-forms *goal*)
                     when (and (is-top-dc-fmla? (formula sf))
			       (dc-index (formula sf)))
                     collect (dc-index (formula sf))))
	(max-index (when indices (apply #'max (mapcar #'symbol-index indices))))
	(fnums
	 (find-all-sformnums 
	  (s-forms *goal*)
	  '*
	  #'(lambda (fmla)
	      (not (and (is-top-dc-fmla? fmla)
			(or (not (dc-index fmla))
			    (equal (symbol-index (dc-index fmla))
				   max-index)))))))
	(x (cons 'delete fnums)))
    x)  " " " ")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  CHECKING FOR AND GETTING THE INTERVAL INDEX ON DC FORMULAS

(defun is-top-dc-fmla? (fmla)
  (cond ((number-expr? fmla) t)
	((variable? fmla) t)
	((constant? fmla) t)
	((not-expr? fmla)
         (is-top-dc-fmla? (args1 fmla)))
;;; special case inserted because the following ifff? crashes if fmla is
;;; a projection-application
	((projection-application? fmla) nil)
;;; special case inserted because the following ifff? crashes if fmla is
;;; a projection-application
	((field-application? fmla) nil)
        ((or (conjunction? fmla)(disjunction? fmla)(implication? fmla)(ifff? fmla))
         (or (is-top-dc-fmla? (args1 fmla))
             (is-top-dc-fmla? (args2 fmla))))
        ((or (exists-expr? fmla) (forall-expr? fmla))
	 (is-top-dc-fmla? (expression fmla)))
	((and (is-compare-symb? fmla)
	      (is-bool? fmla))
	 (or (is-top-dc-fmla? (args1 fmla))
             (is-top-dc-fmla? (args2 fmla))))
	((is-arith-symb? fmla)
	 (or (is-top-dc-fmla? (args1 fmla))
             (is-top-dc-fmla? (args2 fmla))))
	(t (and (application? fmla)
		(singleton? (arguments fmla))
		(type-name? (print-type (type (args1 fmla))))
		(eq (id (print-type (type (args1 fmla)))) '|Interval|)))))

(defun dc-index (fmla)
  (cond ((number-expr? fmla) nil)
	((constant? fmla) nil)
	((variable? fmla) nil)
	((not-expr? fmla)(dc-index (args1 fmla)))
;;; special case inserted because the following ifff? crashes if fmla is
;;; a projection-application
	((projection-application? fmla) nil)
;;; special case inserted because the following ifff? crashes if fmla is
;;; a projection-application
	((field-application? fmla) nil)
        ((or (conjunction? fmla)(disjunction? fmla)(implication? fmla)(ifff? fmla))
         (or (dc-index (args1 fmla))(dc-index (args2 fmla))))
	((and (is-compare-symb? fmla)
	      (is-bool? fmla))
	 (or (dc-index (args1 fmla))
             (dc-index (args2 fmla))))
;;;        ((or (exists-expr? fmla) (all-expr? fmla))
        ((or (exists-expr? fmla) (forall-expr? fmla))
	 (dc-index (expression fmla)))
	((is-arith-symb? fmla)
	 (or (dc-index (args1 fmla))
             (dc-index (args2 fmla))))
	((infix-application? fmla)
	 (or (dc-index (args1 fmla))
             (dc-index (args2 fmla))))
        (t (and (is-dc-indexed? fmla)(args1 fmla)))))


(defun symbol-index (name)
  (let* ((name  (intern  (string  (op-to-id  name))))
	 (string (format nil "~a" name))
	 (pos (position-if #'(lambda (x)(member x (list  #\! #\$))) string))
         (index (when pos (parse-integer string :start (1+ pos)))))
;;; if index is NIL for some reason then return -1
    (or index -1)))


(defun add-index (expr)
  (let* ((indices (loop for sf in (s-forms *goal*)
		      when (and (is-top-dc-fmla? (formula sf))
				(dc-index (formula sf)))
		      collect (dc-index (formula sf))))
         (max-index (when indices (apply #'max (mapcar #'symbol-index indices))))
	 (index (make-dc (format nil "i!~a" max-index) "Interval")))
    (make-application expr index)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  TESTING FUNCTIONS


(defun is-function-of-interval? (fmla)
  (let ((tp (find-supertype (type fmla))))
    (and (funtype? tp)
	 (singleton? (domain tp))
	 (tc-eq (car (domain tp)) (interval-type)))))

(defun is-function-of-time? (fmla)
  (let ((tp (find-supertype (type fmla))))
    (and (funtype? tp)
	 (singleton? (domain tp))
	 (tc-eq (car (domain tp))  (typecheck (pc-parse "Time" 'type-expr))))))

(defun is-bool? (fmla)
  (tc-eq (type fmla) *boolean*))

(defun is-dc-all? (fmla) ;;; Assuming NO outermost dc index
  (and (match-pattern? "fall(a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_quant|)))

(defun is-compare-symb? (fmla)
  (and (infix-application? fmla)
       (not (eq nil (member (id (operator fmla )) (list '< '> '<= '>= '=))))))

(defun is-dc-n? (fmla)
  (and (match-pattern? "n(a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_real|)))

(defun is-dc-dur? (fmla)
  (and (match-pattern? "dur(a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-cf? (fmla)
  (and (match-pattern? "cf(a_)" fmla "a_")
       (or (defined-in? (operator fmla) '|dc_bool|)
	   (defined-in? (operator fmla) '|dc_real|))))

(defun is-arith-symb? (fmla)
  (and (infix-application? fmla)
       (not (eq nil (member (id (operator fmla )) (list '* '+ '- '/))))))

(defun is-dc-indexed? (fmla)
  (and (match-pattern? "a_(a_)" fmla "a_")
       (is-type? (car (select-match "a_(b_)" fmla "b_")) "Interval")))

(defun is-dc-true? (fmla)
  (and (constant? fmla)
       (match-pattern? "TRUE" fmla "a_")
       (defined-in? fmla '|dc_dc|)))

(defun is-dc-false? (fmla)
  (and (constant? fmla)
       (match-pattern? "FALSE" fmla "a_")
       (defined-in? fmla '|dc_dc|)))

(defun is-dc-box? (fmla)
  (and (match-pattern? "[](a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-diamond? (fmla)
  (and (match-pattern? "<>(a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-chop? (fmla)
  (and (match-pattern? (substitute *dc-chop-char* #\; "a_ ; a_") fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-disj? (fmla)
  (and (match-pattern? "a_ \\/ a_" fmla "a_")
       (defined-in? (operator fmla) '|dc_bool|)))

(defun is-dc-conj? (fmla)
  (and (match-pattern? "a_ /\\ a_" fmla "a_")
       (defined-in? (operator fmla) '|dc_bool|)))

(defun is-dc-impl? (fmla)
  (and (match-pattern? "a_ => a_" fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-iff? (fmla)
  (and (match-pattern? "a_ <=> a_" fmla "a_")
       (defined-in? (operator fmla) '|dc_dc|)))

(defun is-dc-neg? (fmla)
  (and (match-pattern? "NOT(a_)" fmla "a_")
       (defined-in? (operator fmla) '|dc_bool|)))

(defun is-dc-tt? (fmla)
  (and (match-pattern? "tt" fmla "a_")
       (defined-in? fmla '|dc_bool|)))

(defun is-dc-ff? (fmla)
  (and (match-pattern? "ff" fmla "a_")
       (defined-in? fmla '|dc_bool|)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  CONSTRUCTOR FUNCTIONS


;;; Create an operator with name opstr (string) and the type given
;;; in typestr (string)

(defun make-dc (opstr typestr)
  (let* ((op (pc-parse opstr 'expr)))
    (typecheck op :expected
	       (typecheck (pc-parse typestr 'type-expr)))))


;;; Create an application operator with name opstr (string) and the 
;;; type given typestr (string) on the expression exp

(defun make-dc-application (opstr typestr exp)
  (let* ((op (pc-parse opstr 'expr))
	 (res (mk-application op (car exp))))
    (typecheck res :expected
	       (typecheck (pc-parse typestr 'type-expr)))))


;;; Create a name expression with name opstr (string) and the 
;;; type given typestr (string) 

(defun make-dc-name-expr (opstr typestr)
  (let* ((obj (pc-parse opstr 'expr)))
;;;	 (res (mk-name-expr obj)))
    (typecheck obj :expected
	       (typecheck (pc-parse typestr 'type-expr)))))

;;; Create an expression with name opstr (string) and the 
;;; type given typestr (string) on a list of expressions exprs.  The base
;;; element is basestr.

(defun make-dc-rec (opstr basestr typestr exprs)
  (let* ((op (pc-parse opstr 'expr))
	 (base (pc-parse basestr 'expr))
	 (conjunction (mk-rec-application op base exprs)))
    (typecheck conjunction :expected
	       (typecheck (pc-parse typestr 'type-expr)))))

(defun make-dc-lambda (vars exp)
  (let* ((lamexp (mk-lambda-expr vars exp)))
    (typecheck lamexp :expected
	       (typecheck (pc-parse "[real -> [Interval -> bool]]" 'type-expr)))))

(defun make-dc-compare (symb exprs)
  (let* ((res (mk-application symb (car exprs) (cadr exprs))))
    (typecheck res :expected
	       (typecheck (pc-parse "IForm" 'type-expr)))))

(defun make-dc-arith (symb exprs)
  (let* ((res (mk-application symb (car exprs) (cadr exprs))))
    (typecheck res :expected
	       (typecheck (pc-parse "ITime" 'type-expr)))))

(defun make-bool-compare (symb exprs)
  (let* ((res (mk-application symb (car exprs) (cadr exprs))))
    (typecheck res :expected
	       (typecheck (pc-parse "bool" 'type-expr)))))

(defun make-Time-arith (symb exprs)
  (let* ((res (mk-application symb (car exprs) (cadr exprs))))
    (typecheck res :expected
	       (typecheck (pc-parse "Time" 'type-expr)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  CONVERTING FROM IFORM TO BOOL


(defun convert-from-dc (expr)
  (cond ((is-dc-n? expr) (args1 expr))
        ((and (exists-expr? expr) (is-type? (car (bindings expr)) "real"))
	 (make-application (make-dc "ex" "[[real -> IForm] -> IForm]")
			   (make-dc-lambda (bindings expr)
					   (convert-from-dc (expression expr)))))
        ((and (exists-expr? expr) (is-type? (car (bindings expr)) "nzreal"))
	 (make-application (make-dc "ex" "[[nzreal -> IForm] -> IForm]")
			   (make-dc-lambda (bindings expr)
					   (convert-from-dc (expression expr)))))
;;;        ((and (all-expr? expr) (is-type? (car (bindings expr)) "real"))
        ((and (forall-expr? expr) (is-type? (car (bindings expr)) "real"))
	 (make-application (make-dc "fall" "[[real -> IForm] -> IForm]")
			   (make-dc-lambda (bindings expr)
					   (convert-from-dc (expression expr)))))
;;;        ((and (all-expr? expr) (is-type? (car (bindings expr)) "nzreal"))
        ((and (forall-expr? expr) (is-type? (car (bindings expr)) "nzreal"))
	 (make-application (make-dc "fall" "[[nzreal -> IForm] -> IForm]")
			   (make-dc-lambda (bindings expr)
					   (convert-from-dc (expression expr)))))
        ((is-dc-conj? expr) 
         (make-conjunction
	  (mapcar #'convert-from-dc (arguments expr))))
        ((is-dc-disj? expr)
	 (make-disjunction
	  (mapcar #'convert-from-dc (arguments expr))))
        ((is-dc-impl? expr)
	 (make-implication
	  (mapcar #'convert-from-dc (arguments expr))))
        ((is-dc-iff? expr)
	 (make-iff
	  (mapcar #'convert-from-dc (arguments expr))))
        ((is-dc-neg? expr)
	 (make-dc-application "not" "IForm" (convert-from-dc (args1 expr))))
	((is-compare-symb? expr)
	 (make-bool-compare (id (operator expr))
			    (mapcar #'convert-from-dc (arguments expr))))
	((is-arith-symb? expr)
	 (make-Time-arith (id (operator expr))
			  (mapcar #'convert-from-dc (arguments expr))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  RETRIEVING FORMULAS FROM THE GOAL SEQUENT

;;; Return the formula number (in a list) of a formula satisfying pred or
;;; nil if no such formula is found.
;;; Searches in the intersection of formulas designated by sformnum and nums
;;; sformnum may be a number, +, - or *.
;;; nums may be a +, -, * or a list of numbers
(defun find-dc-num (sformnum nums pred)
  (list (find-sform (select-seq (s-forms *goal*) nums) sformnum
		    #'(lambda (sform)
			(if (not-expr? (formula sform))
			    (and (is-dc-indexed? (args1 (formula sform)))
				 (funcall pred (operator (args1 (formula sform)))))
			  (and (is-dc-indexed? (formula sform))
			       (funcall pred (operator (formula sform)))))))))

(defun find-dc-pattern (sformnum pattern nums)
  (let ((patt (substitute *dc-chop-char* #\; pattern)))
    (find-dc-num sformnum nums #'(lambda (x) (match-pattern? patt x "a_")))))

(defun find-all-dc-nums (sformnum nums pred)
  (find-all-sformnums (select-seq (s-forms *goal*) nums) sformnum
		      #'(lambda (sform)
			  (if (not-expr? sform)
			      (and (is-dc-indexed? (args1 sform))
				   (funcall pred (operator (args1 sform))))
			    (and (is-dc-indexed? sform)
				 (funcall pred (operator sform)))))))

(defun find-all-dc-pattern (sformnum pattern nums)
  (find-all-dc-nums sformnum nums #'(lambda (x) (match-pattern? pattern x "a_"))))

(defun find-dc-fmla (nums)
  (let ((fmla (formula (car (select-seq (s-forms *goal*) nums)))))
    (if (not-expr? fmla)
	(operator (args1 fmla))
      (operator fmla))))

(defun find-pvs-fmla (nums)
  (let ((fmla (formula (car (select-seq (s-forms *goal*) nums)))))
    (if (not-expr? fmla)
	(args1 fmla)
      fmla)))

(defun find-pvs-num (sformnum nums pred)
  (list (find-sform (select-seq (s-forms *goal*) nums) sformnum
		    #'(lambda (sform)
			(if (not-expr? (formula sform))
			    (funcall pred (args1 (formula sform)))
			  (funcall pred (formula sform)))))))

;;; Use DC syntax in the pattern
(defun find-pvs-pattern (sformnum pattern nums)
  (let ((patt (substitute *dc-chop-char* #\; pattern)))
    (find-pvs-num sformnum nums #'(lambda (x) (match-pattern? patt x "a_")))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  SELECTOR FUNCTIONS ON EXPRESSIONS 

(defun formulas (sforms) (mapcar 'formula sforms))

(defun dc-args (fmls) (mapcar 'args1 fmls))

(defun dc-operators (fmls) (mapcar 'operator fmls))

(defun delete-dc-seq (nums)
  (delete-seq (s-forms *goal*) nums))

(defun gather-dc-seq (yesnums nonums)
  (gather-seq (s-forms *goal*) yesnums nonums))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  MAKING DURATION CALCULUS CONJUNCT AND DISJUNCT FROM GOAL SEQUENT

;;; abstract term in form by var
(defun abstract (form term var)
  ;; assume term typechecked
  (let* ((var (pc-parse var 'name))
	 (type (type term))
	 (var-decl (make-bind-decl (id var) type))
	 (var (make-variable-expr var-decl))
	 (newform (termsubst form
			     #'(lambda (x) var)
			     #'(lambda (x) (tc-eq x term))))
					;	 (newform (universal-closure newform))
					;	 (freevars-form  (freevars newform))
	 (freevars-form  (list var))
	 (newform 
	  (close-freevars newform *current-context* freevars-form nil t))
	 (newform (quant-to-lambda newform)))
    newform))

(defun abstract-interval (form interval)
  (abstract form interval (gentemp "i_")))


(defun collect-fnums (seq &optional (fnums *))
  (let* ((sforms (select-seq seq fnums))
	 (antec-fmlas
	  (loop for sf in sforms when (negation? (formula sf))
	      collect (args1 (formula sf))))
	 (conseq-fmlas
	  (loop for sf in sforms when (not (negation? (formula sf)))
	      collect (formula sf)))
	 (fmla (make-implication
		(make-conjunction antec-fmlas)
		(make-disjunction conseq-fmlas))))
    fmla))

(defun collect-antec (seq)
  (let* ((sforms (select-seq seq '-))
	 (antec-fmlas
	  (loop for sf in sforms when (negation? (formula sf))
	      collect (args1 (formula sf))))
	 (fmla (make-conjunction antec-fmlas)))
    fmla))

(defun collect-conseq (seq)
  (let* ((sforms (select-seq seq '+))
	 (conseq-fmlas
	  (loop for sf in sforms when (not (negation? (formula sf)))
	      collect (formula sf)))
	 (fmla (make-disjunction conseq-fmlas)))
    fmla))

(defun term-occures-in? (term fmla)
  (let ((pred #'(lambda (x) (tc-eq x term))))
    (collect-subterms fmla pred)))

;;; This predicates defines which formulas are considered
;;; dc formulas. Currently if the interval term occures in fmla then 
;;; fmla is a dc formula
(defun is-dc-fmla? (interval fmla)
  (term-occures-in? interval fmla))

;;; Returns a conjunction of those antecedent formulas of seq
;;; which are considered dc formulas
(defun conjunc-dc-antec (i seq)
  (let* ((sforms (select-seq seq '-))
	 (antec-fmlas
	  (loop for sf in sforms when 
		(and (negation? (formula sf))
		     (is-dc-fmla? i (formula sf)))
	      collect (args1 (formula sf))))
	 (fmla (make-conjunction antec-fmlas)))
    fmla))

;;; Returns a disjunction of those consequent formulas of seq
;;; which are considered dc formulas
(defun disjunc-dc-conseq (i seq)
  (let* ((sforms (select-seq seq '+))
	 (conseq-fmlas
	  (loop for sf in sforms when 
		(and (not (negation? (formula sf)))
		     (is-dc-fmla? i (formula sf)))
	      collect (formula sf)))
	 (fmla (make-disjunction conseq-fmlas)))
    fmla))

(defun collect-dc-antec (interval seq)
  (let* ((antec (conjunc-dc-antec interval seq))
	 (dc-antec (abstract-interval antec interval)))
    dc-antec))

(defun collect-dc-conseq (interval seq)
  (let* ((conseq (disjunc-dc-conseq interval seq))
	 (dc-conseq (abstract-interval conseq interval)))
    dc-conseq))


;;; Analoguous to hide-index. But only deletes DC formulas indexed with i.
;;; hide-index deletes ANY formulas which is not indexed with the largest
;;; index.
(defstep delete-dc-fmlas (i)
  (let ((fnums
	 (find-all-sformnums 
	  (s-forms *goal*)
	  '*
	  #'(lambda (fmla)
	      (is-dc-fmla? i fmla))))
	(x (cons 'delete fnums)))
    x)  " " " ")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  FIND DC CONTEXT INTERVAL FROM GOAL

(defun interval-constant? (expr)
  (and (name-expr? expr) 
       (skolem-const-decl? (declaration expr))
       (is-type? expr "Interval")))

(defun find-dc-interval ()
  (let* ((fmlas (collect-fnums (s-forms *goal*)))
	 (interval-constants 
	  (collect-subterms fmlas #'interval-constant?))
	 (interval-constants
	  (sort interval-constants
		#'(lambda (x y)
		    (member x (collect-subterms y #'interval-constant?))))))
    (when (not (null interval-constants))
      (car interval-constants))))

(defun find-dc-context-interval (&optional ist)
  (if ist 
      (pc-parse-dc-interval ist)
    (find-dc-interval)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  SPECIAL FUNCTIONS FOR DUR-PRESERVE PROOF RULE

(defun new-var (prefix type)
  (generate-variable (gentemp prefix) type *current-context*))

(defun get-states (expr)
  (cond ((is-dc-n? expr) nil)
	((is-dc-dur? expr) (list expr))
        ((infix-application? expr)
         (nconc (get-states (args1 expr))
		(get-states (args2 expr))))
        ((unary-application? expr)
	 (get-states (args1 expr)))
	(t nil)))

(defun insert-dc (exp dur subst)
  (termsubst exp #'(lambda (x) subst) #'(lambda (x) (tc-eq x dur))))

(defun make-lambda-from-expr-iform (expr)
  (let* ((states (get-states expr))
	 (v1 (new-var "ir_" (IReal-type)))
	 (v2 (new-var "is_" (IReal-type)))
	 (expr2 (insert-dc (insert-dc expr (car states) v1)
			   (cadr states) v2))
	 (lmbd (quant-to-lambda
		(close-freevars (typecheck expr2 :expected (IForm-type))
				*current-context*
				(list v1 v2)))))
    lmbd))

(defun make-lambda-from-expr-bool (expr)
  (let* ((states (get-states expr))
	 (v1 (new-var "r_" (dc-real-type)))
	 (v2 (new-var "s_" (dc-real-type)))
	 (nv1 (typecheck (mk-application '|n| v1)
			 :expected (IReal-type)))
	 (nv2 (typecheck (mk-application '|n| v2)
			 :expected (IReal-type)))
	 )
    (quant-to-lambda
     (close-freevars
      (convert-from-dc
       (typecheck (insert-dc (insert-dc expr (car states) nv1)
			     (cadr states) nv2) :expected (IForm-type)))
      *current-context*
      (list v1 v2)))))

(defun make-interval-type (st)
  (typecheck (pc-parse (format nil "[Interval -> ~a]" st) 'type-expr)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  PARSING AND TYPECHECKING FUNCTION 

(defun gen-pc-parse (typestr exp)
  (let* ((op (pc-parse exp 'expr)))
    (typecheck op :expected
	       (typecheck (pc-parse typestr 'type-expr)))))

(defun pc-parse-dc-interval (exp)
  (let* ((op (pc-parse exp 'expr)))
    (typecheck op :expected Interval-type)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  CONTEXT FREENESS 

(defun is-dc-context-free? (expr)
  (cond ((is-dc-n? expr) t)
        ((or (is-dc-conj? expr)(is-dc-disj? expr)
	     (is-dc-impl? expr)(is-dc-iff? expr))
         (and (is-dc-context-free? (args1 expr))
	      (is-dc-context-free? (args2 expr))))
        ((is-dc-neg? expr)
	 (is-dc-context-free? (args1 expr)))
	((is-compare-symb? expr)
	 (and (is-dc-context-free? (args1 expr))
	      (is-dc-context-free? (args2 expr))))
	((is-arith-symb? expr)
	 (and (is-dc-context-free? (args1 expr))
	      (is-dc-context-free? (args2 expr))))
	(t nil)))

(defun find-context-free-conj (expr)
  (cond ((is-dc-cf? expr) expr)
        ((is-dc-conj? expr)
         (or (find-context-free-conj (args1 expr))
             (find-context-free-conj (args2 expr))))
	(t nil)))

(defun gather-not-context-free (expr cf-expr)
  (cond ((eq expr cf-expr) (make-dc-name-expr "TRUE" "IForm"))
        ((and (is-dc-conj? expr) (eq (args1 expr) cf-expr))
	 (args2 expr))
        ((and (is-dc-conj? expr) (eq (args2 expr) cf-expr))
	 (args1 expr))
        ((is-dc-conj? expr)
	 (make-dc-rec "/\\" "TRUE" "IForm"
		      (mapcar `(lambda(x)(gather-not-context-free x ,cf-expr))
			      (arguments expr))))
	(t expr)))
