(***************************************************************************
 File "bdd.sml" Updated 1993-02-09
 Duration Calculus checker, version 2
 Applicative binary decision diagrams for transitions
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature BDD =
sig 
    eqtype 'a Bdd
    val mapbdd: ('a -> 'b) -> 'a Bdd -> 'b Bdd
    val combinebdd: ('a * 'b -> 'c) -> 'a Bdd * 'b Bdd -> 'c Bdd
    val range: 'a Bdd -> 'a list
    val comparebdd: ('a -> ''b) -> 'a Bdd * 'a Bdd -> bool
    val reduce: ''a Bdd -> ''a Bdd
    val buildbdd: string list * string list list * 'a * 'a -> 'a Bdd
    val bdd2assoc: 'a Bdd * string list -> (string * 'a) list
end

functor Bdd(structure Basic: BASIC): BDD =
struct
    open Basic
    datatype 'a Bdd =			
	Lf of 'a
      | FT of 'a Bdd * 'a Bdd
      | Both of 'a Bdd 
	
(* Idea: (Both t) abbreviates FT(t,t).  A BDD is *reduced* if it has
no node FT(t1,t2) with t1 = t2.  The operations below do not
necessarily yield reduced BDDs. *)

(* Map function f over all leaves of the BDD t *)

fun mapbdd f t =
    let fun h (Lf b)      = Lf (f b)
	  | h (FT(t1,t2)) = FT(h t1, h t2)
	  | h (Both t)    = Both(h t)
    in h t end

(* Combine two BDDs using the leaf combiner f. *)

fun combinebdd f =
    let fun h (Lf b1,       Lf b2)       = Lf( f(b1,b2) )
	  | h (FT(t11,t12), Both t2)     = FT(h(t11,t2),  h(t12,t2))
	  | h (FT(t11,t12), FT(t21,t22)) = FT(h(t11,t21), h(t12,t22))
	  | h (Both t1,     Both t2)     = Both(h(t1,t2))
	  | h (Both t1,     FT(t21,t22)) = FT(h(t1,t21), h(t1,t22))
    in h end
		  
(* Compute a list of all elements of a transition BDD t *)

fun range (Lf b)      = [b]
  | range (FT(t1,t2)) = range(t1)@range(t2)
  | range (Both t)    = range(t)

(* Compare transition BDDs t1 and t2 wrt. partition of the states.
Returns true if on every symbol both lead to states belonging to the
same equivalence class.  The partition is represented by function
classof such that classof a = classof b precisely when a and b are in
the same class.  *)

fun comparebdd classof (t1, t2) =
    let fun h (Lf b1, Lf b2)             = (classof b1 = classof b2)
	  | h (FT(t11,t12), Both t2)     = h(t11, t2) andalso h(t12,t2)
	  | h (FT(t11,t12), FT(t21,t22)) = h(t11, t21) andalso h(t12,t22)
	  | h (Both t1, FT(t21,t22))     = h(t1, t21) andalso h(t1,t22)
	  | h (Both t1, Both t2)         = h(t1, t2) 
    in h(t1, t2) end

(* Reduce the BDD, ie. replace FT(t,t) by (Both t) throughout *)

fun reduce (Lf a)       = Lf a
  | reduce (Both t)     = Both (reduce t)
  | reduce (FT(t1, t2)) =
    let val t1' = reduce t1
	val t2' = reduce t2
    in if t1' = t2' then Both t1' else FT(t1', t2') end

(* Build the BDD corresponding to a logical expression in disjunctive
normal form.  The DNF is given as a (disjunctive) list "conjuncts" of
(conjunctive) lists of state variables "p" or negated state variables
"~p".  In the DNF, [] <=> false (empty disjunction), and [[]] <=> true
(empty conjunction).

Justification of the algorithm below: In the "true" branch of "p",
every conjunct containing "~p" is unsatisfiable and can therefore be
deleted (using: false /\ Q <=> false) (and dually for the "false"
branch).  Moreover, in the "true" branch, "p" is everywhere "true" and
can be deleted from all conjuncts in which it appears (using: true /\
Q <=> Q).  If some conjunct becomes empty, then it is always true, and
all other conjuncts in the BDD branch can be deleted (using: true \/ Q
<=> true).  If the "true" and "false" branches of a node have the same
(disjunctive) set of conjuncts, then their truth values will be
identical, so the node is a "Both".  All leaves with an empty
disjunction are false (and get decorated with fal); all leaves with a
non-empty disjunction are true (and get decorated with tru).

If each of the sublists representing conjuncts is sorted, then a
reduced BDD (without FT(t,t)) will be generated.  *)

fun buildbdd (snames, conjuncts, fal, tru) = 
    let fun delete p bcs = 
	    let val reducedbcs = map (filter (fn x => x<>p)) bcs
	    in if mem [] reducedbcs then [[]] else reducedbcs end
	fun h ([], [])    = Lf fal
	  | h ([], _)     = Lf tru
	  | h (p::r, bcs) = 
	    let val bcsleft  = delete ("~"^p) (filter (not o (mem p)) bcs)
		val bcsright = delete p (filter (not o (mem ("~"^p))) bcs)
	    in if bcsleft = bcsright then
		   Both (h(r, bcsleft))
	       else
		   FT(h(r, bcsleft), h(r, bcsright))
	    end
    in h(snames, conjuncts) end

    fun bdd2assoc (t, snames) =
	let fun b2a (Lf a, _, p, res) = (p, a) :: res
	      | b2a (Both t, sn::snr, p, res) = b2a(t, snr, p, res)
	      | b2a (FT(t1, t2), sn::snr, p, res) =
		b2a(t2, snr, p ^ sn, b2a(t1, snr, p ^ "~" ^ sn, res))
	in b2a(t, snames, "", []) end
end
