(***************************************************************************
 File "formula.sml" Updated 1993-02-08
 Duration Calculus checker, version 2
 Scanning, parsing, simplifying, and printing DC formulas
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature FORMULA =
sig
    datatype State = 
	Id of string
      | tt
      | ff
      | sNeg of State
      | sAnd of State * State
      | sOr of State * State
      | sIff of State * State
      | sImp of State * State
	
    datatype Formula = 
	True
      | False 
      | Empty
      | Box of Formula
      | Diamond of Formula
      | Neg of Formula
      | Chop of Formula * Formula
      | And of Formula * Formula
      | Or of Formula * Formula
      | Iff of Formula * Formula
      | Imp of Formula * Formula
      | Pred of State

    exception Scanerror of string
    exception Parseerror of string

    val parse: string -> Formula
    val simplify: Formula -> Formula
    val statenames: Formula -> string list
    val basic_conjuncts: State -> string list list
    val prtformula: Formula -> string
end   

functor Formula (structure Basic: BASIC): FORMULA =
struct
    open Basic

    datatype State = 
	Id of string
      | tt
      | ff
      | sNeg of State
      | sAnd of State * State
      | sOr of State * State
      | sIff of State * State
      | sImp of State * State
	
    datatype Formula = 
	True
      | False 
      | Empty
      | Box of Formula
      | Diamond of Formula
      | Neg of Formula
      | Chop of Formula * Formula
      | And of Formula * Formula
      | Or of Formula * Formula
      | Iff of Formula * Formula
      | Imp of Formula * Formula
      | Pred of State
 
    (* JUS 1992-09-30 Scanning and parsing for PDC. *)

    (* SCANNING to a list of terminal symbols **************************)

    datatype terminal =
	Lpar' | Rpar' | Box' | Diamond' | Ldash' | Rdash'
      | And' | Or' | Neg' | Iff' | Imp' | Semi' 
      | Name' of string 
    exception Scanerror of string

    fun isblank c = (c = " " orelse c = "\t" orelse c = "\n")
    fun isdigit c = ("0" <= c andalso c <= "9")
    fun islower c = ("a" <= c andalso c <= "z")
    fun isupper c = ("A" <= c andalso c <= "Z")
    fun isletter c = (islower c orelse isupper c)
    fun isletterdigit c = (isletter c orelse isdigit c)
	
    fun readfile (filename: string): string =
        let val infile = open_in filename 
            val str    = input (infile, can_input infile)
            val ()     = close_in infile
        in str end
    
    fun scname (cs, value) =
	case cs of 
	    c :: cr => if isletterdigit c then scname(cr, c :: value)
		       else (cs, implode (rev value))
	  | []      => (cs, implode (rev value))
		
    fun scan s =
	let fun sc cs = 
	    case cs of
		[]             => []
	      | "(" :: cr      => Lpar' :: sc cr     
	      | ")" :: cr      => Rpar' :: sc cr     
	      | "[" :: "]" :: cr      => Box' :: sc cr     
	      | "<" :: ">" :: cr      => Diamond' :: sc cr     
	      | "'" :: cr      => Ldash' :: sc cr     
  (*          | "`" :: cr      => Rdash' :: sc cr     *)
	      | "&" :: cr      => And' :: sc cr     
	      | "|" :: cr      => Or' :: sc cr     
	      | "~" :: cr      => Neg' :: sc cr     
	      | "<" :: "=" :: ">" :: cr      => Iff' :: sc cr     
	      | "=" :: ">" :: cr      => Imp' :: sc cr     
	      | ";" :: cr 	   => Semi' :: sc cr
	      | c :: cr        => if isblank c then sc cr
				  else if isletter c then
				      let val (cs1, n) = scname(cr, [c])
				      in Name' n :: sc cs1 end
				  else raise Scanerror (implode cs)
	in sc (explode s) end


    (* SH 1997-07-24 unscan added to provide error information in exceptions *)

    fun unscan ts =
      let fun unscan_terminal t =
        case t of
            Lpar' => "("
          | Rpar' => ")"
          | Box' => "[]"
          | Diamond' => "<>"
          | Ldash' => "'"
          | Rdash' => "'"
          | And' => "/\\"
          | Or' => "\\/"
          | Neg' => "NOT"
          | Iff' => "<=>"
          | Imp' => "=>"
          | Semi' => ";" 
          | Name' s => s
      in
        case ts of
            [] => ""
          | t::ts => (unscan_terminal t)^(unscan ts)
      end

    (* JUS 1992-09-30 Parsing for PDC. *)
	
    (* PARSING to  abstract syntax ****************************************)
	
    (* Grammar, non left recursive and left factorized 
     
 F      = 'true' F1 
        | 'false' F1
	| '[]' F F1
	| '<>' F F1
	| '~' F F1
	| ''' F2
	| '(' F ')' F1 .
	  
 F1     = ';' F F1
        | '&' F F1
        | '|' F F1
        | '<=>' F F1
        | '=>' F F1
	| empty .

 F2     = Id S1 ''' F1
        | 'tt' S1 ''' F1
        | 'ff' S1 ''' F1
        | '~' S S1 ''' F1
	| '(' S ')' S1 ''' F1
        | ''' F1

 S      = Id S1
        | 'tt' S1
        | 'ff' S1
        | '~' S S1
	| '(' S ')' S1

 S1     = '&' S S1
        | '&' S S1
        | '<=>' S S1
        | '=>' S S1
	| empty .
*)


    exception Parseerror of string

    fun Formula ts =
	case ts of
	    Name' "true" :: tr   => Formula1 (tr, True)
	  | Name' "false" :: tr  => Formula1 (tr, False)
	  | Ldash' :: tr         => let val (ts1, tv) = Formula2 tr 
				    in (ts1, tv) end
	  | Box' :: tr           => let val (ts1, tv) = Formula tr 
					val (ts2, ev) = Formula1 (ts1, Box tv)
				    in (ts2, ev) end
	  | Diamond' :: tr       => let val (ts1, tv) = Formula tr 
					val (ts2, ev) = Formula1 (ts1, Diamond tv)
				    in (ts2, ev) end
	  | Neg' :: tr           => let val (ts1, tv) = Formula tr 
					val (ts2, ev) = Formula1 (ts1, Neg tv)
				    in (ts2, ev) end
	  | Lpar' :: tr          => let val (ts1, tv) = Formula tr 
				    in case ts1 of
					Rpar' :: ts2 => let val (ts3,ev) = 
					    Formula1(ts2,tv)
							in (ts3, ev) end
				      | _           => raise Parseerror (unscan ts1)
				    end
	  | _                   => raise Parseerror (unscan ts)
		
    and Formula1 (ts, inval) =
	case ts of
	    Semi' :: tr => let val (ts1, tv) = Formula tr 
			       val (ts2, ev) = Formula1 (ts1, Chop(inval, tv))
			   in (ts2, ev) end
	  | And' :: tr  => let val (ts1, tv) = Formula tr 
			       val (ts2, ev) = Formula1 (ts1, And(inval, tv))
			   in (ts2, ev) end
	  | Or' :: tr   => let val (ts1, tv) = Formula tr 
			       val (ts2, ev) = Formula1 (ts1, Or(inval, tv))
			   in (ts2, ev) end
	  | Iff' :: tr  => let val (ts1, tv) = Formula tr 
			       val (ts2, ev) = Formula1 (ts1, Iff(inval, tv))
			   in (ts2, ev) end
	  | Imp' :: tr  => let val (ts1, tv) = Formula tr 
			       val (ts2, ev) = Formula1 (ts1, Imp(inval, tv))
			   in (ts2, ev) end
	  | _          => (ts,inval)
		
    and Formula2 ts =
	case ts of
	    Ldash' :: tr => Formula1(tr, Empty)
	  (*        Rdash' :: tr => Formula1(tr, Empty) *)
	  | t :: tr     => let in
		case t of
		    Name' "tt" => let val (ts1, tv) = State ts
				  in case ts1 of
				      Ldash' :: ts2 => Formula1(ts2,
								Pred(tv))
				    | _           => raise Parseerror (unscan ts1)
				  end
		  | Name' "ff" => let val (ts1, tv) = State ts
				  in case ts1 of
				      Ldash' :: ts2 => Formula1(ts2,Pred(tv))
				    | _           => raise Parseerror (unscan ts1)
				  end
		  | Name' n  => let val (ts1, tv) = State ts
				in case ts1 of
				    Ldash' :: ts2 => Formula1(ts2,Pred(tv))
				  | _           => raise Parseerror (unscan ts1)
				end
		  | Neg'  => let val (ts1, tv) = State ts
			     in case ts1 of
				 Ldash' :: ts2 => Formula1(ts2,Pred(tv))
			       | _           => raise Parseerror (unscan ts1)
			     end
		  | Lpar'  => let val (ts1, tv) = State ts
			      in case ts1 of
				  Ldash' :: ts2 => Formula1(ts2,Pred(tv))
				| _           => raise Parseerror (unscan ts1)
			      end
		  | _          => raise Parseerror (unscan ts)
			   end
	  | _          => raise Parseerror (unscan ts)
		
    and State ts =
	case ts of
	    Name' "tt" :: tr     => State1 (tr, tt)
	  | Name' "ff" :: tr     => State1 (tr, ff)
	  | Name' n :: tr        => State1 (tr, Id(n))
	  | Neg' :: tr           => let val (ts1, tv) = State tr 
					val (ts2, ev) = State1 (ts1, sNeg tv)
				    in (ts2, ev) end
	  | Lpar' :: tr          => let val (ts1, tv) = State tr 
				    in case ts1 of
					Rpar' :: ts2 => let val (ts3,ev) = 
					    State1(ts2,tv)
							in (ts3, ev) end
				      | _           => raise Parseerror (unscan ts1)
				    end
	  | _                   => raise Parseerror (unscan ts)
		
    and State1 (ts, inval) =
	case ts of
	    And' :: tr  => let val (ts1, tv) = State tr 
			       val (ts2, ev) = State1 (ts1, sAnd(inval, tv))
			   in (ts2, ev) end
	  | Or' :: tr   => let val (ts1, tv) = State tr 
			       val (ts2, ev) = State1 (ts1, sOr(inval, tv))
		      in (ts2, ev) end
	  | Iff' :: tr  => let val (ts1, tv) = State tr 
			       val (ts2, ev) = State1 (ts1, sIff(inval, tv))
			   in (ts2, ev) end
	  | Imp' :: tr  => let val (ts1, tv) = State tr 
			       val (ts2, ev) = State1 (ts1, sImp(inval, tv))
			   in (ts2, ev) end
	  | _          => (ts,inval)
	
    fun parse s = 
	case (Formula (scan s)) of 
	    ([], result) => result
	  | (ts, _)  => raise Parseerror (unscan ts)
	
(*========================================================================
=== FUNCTION: NForm
=== PURPOSE:  Normalizes a state expression to a disj.-conj.-neg. normal
===           form.      
=== EFFECT:   Given a state expression, the state expression is transformed
===           into a normal form. 
========================================================================*)
fun NForm s =
    let val nf = NForm1(s) in
	if nf = s then s else NForm(nf)
    end

and NForm1 s =
  case s of
     Id(st) => Id(st)
    | tt => tt
    | ff => ff
    | sNeg(sAnd(s1,s2)) => sOr(NForm1(sNeg(s1)),NForm1(sNeg(s2)))
    | sNeg(sOr(s1,s2)) => sAnd(NForm1(sNeg(s1)),NForm1(sNeg(s2)))
    | sNeg(tt) => ff
    | sNeg(ff) => tt
(* SH 1997-707-25 Added to ensure that double negation gets normalized. *)
    | sNeg(sNeg(s1)) => NForm1(s1)
    | sNeg(s1) => sNeg(NForm1(s1))
    | sAnd(sOr(s1,s2),s3) => NForm1(sOr(sAnd(s1,s3),sAnd(s2,s3)))
    | sAnd(s1,sOr(s2,s3)) => NForm1(sOr(sAnd(s1,s2),sAnd(s1,s3)))
    | sAnd(s1,tt) => NForm1(s1)
    | sAnd(tt,s1) => NForm1(s1)
    | sAnd(s1,ff) => ff
    | sAnd(ff,s1) => ff
    | sAnd(s1,s2) => sAnd(NForm1(s1),NForm1(s2))
    | sOr(s1,ff) => NForm1(s1)
    | sOr(ff,s1) => NForm1(s1)
    | sOr(s1,tt) => tt
    | sOr(tt,s1) => tt
    | sOr(s1,s2) => sOr(NForm1(s1),NForm1(s2))
    | sIff(s1,s2) => NForm1(sAnd(sImp(s1,s2),sImp(s2,s1)))
    | sImp(s1,s2) => NForm1(sOr(sNeg(s1),s2))

(***************************************************************************
                      S I M P L I F I C A T I O N
***************************************************************************)

(*========================================================================
=== FUNCTION: simplify
=== PURPOSE:  Transforms a formula to a restricted formula only involving
===           negation, disjunction and chop, and normalizes the states.
=========================================================================*)

fun simplify f =
    let val nf = simp1(f) in
	if nf = f then f else simplify nf
    end

and simp1 f =
  case f of 
      True        => Neg(Pred(ff))
    | False       => Pred(ff)
    | Empty       => Neg(Pred(tt))
    | Box(f1)     => simp1(Neg(Diamond(Neg(f1))))
    | Diamond(f1) => simp1(Chop(True,Chop(f1,True)))
    | Neg(f1)     => Neg(simp1(f1))
    | Chop(f1,f2) => Chop(simp1(f1),simp1(f2))
    | And(f1,f2)  => Neg(Or(Neg(simp1(f1)),Neg(simp1(f2))))
    | Or(f1,f2)   => Or(simp1(f1),simp1(f2))
    | Iff(f1,f2)  => simp1(And(Imp(f1,f2),Imp(f2,f1)))
    | Imp(f1,f2)  => Or(Neg(simp1(f1)),simp1(f2))
    | Pred(s)     => Pred(NForm s)

(*========================================================================
=== FUNCTION: statenames
=== PURPOSE:  Produces a sorted list of all state variables.
=== EFFECT:   Given a formula, it returns a list of all state variables 
===           occurring in the formula (in positive or negative form).
========================================================================*)

fun statenames f = sortstr(setof(states1(f)))

and states1 f =
   case f of
      True        => []
    | False       => []
    | Empty       => []
    | Box(f1)     => states1(f1)
    | Diamond(f1) => states1(f1)
    | Neg(f1)     => states1(f1)
    | Chop(f1,f2) => states1(f1)@states1(f2)
    | And(f1,f2)  => states1(f1)@states1(f2)
    | Or(f1,f2)   => states1(f1)@states1(f2)
    | Iff(f1,f2)  => states1(f1)@states1(f2)
    | Imp(f1,f2)  => states1(f1)@states1(f2)
    | Pred(s)     => statess(s)

and statess s =
  case s of
     Id(st)       => [st]
    | tt          => []
    | ff          => []
    | sNeg(s1)    => statess(s1)
    | sAnd(s1,s2) => statess(s1)@statess(s2)
    | sOr(s1,s2)  => statess(s1)@statess(s2)
    | sIff(s1,s2) => statess(s1)@statess(s2)
    | sImp(s1,s2) => statess(s1)@statess(s2)

(*========================================================================
=== FUNCTION:    basic_conjuncts
=== PURPOSE:     Determines the conjuncts corresponding to a state exp.
========================================================================*)

(* SH 1997-07-25 Changed to satisfy precondition of basic_and and 
   thus avoid Match exception.

fun basic_conjuncts(s:State) =
    case s of
	tt => [[]]
      | ff => []
      | _  => basic_conjuncts1 s 
*)

fun basic_conjuncts(s:State) =
    let val s = (NForm s)
    in
       case s of
  	   tt => [[]]
         | ff => []
         | _  => basic_conjuncts1 s
    end

and basic_conjuncts1(s:State) =
    case s of
        Id(st)          => [basic_and(s)]
      | sNeg(Id(st))    => [basic_and(s)]
      | sOr(s1,s2)      => basic_conjuncts1(s1)@basic_conjuncts1(s2)
      | sAnd(s1,s2)     => [basic_and(s)]

(*========================================================================
=== FUNCTION:    basic_and
=== PURPOSE:     Produces a conjunct based on a state expression.
=== PRECOND.:    The state expression is a conjunction of (possibly negated)
===              state variables. That is, it must be normalized using NForm.
========================================================================*)

and basic_and(s:State) =
    case s of
        Id(st)          => [st]
      | sNeg(Id(st))    => ["~"^st]
      | sAnd(s1,s2)     => basic_and(s1)@basic_and(s2)
     
(***************************************************************************
                      P R I N T O U T   O F   F O R M U L A
 ***************************************************************************)

    fun prtformula f =
	let fun pp1 f =
	    case f of 
		True => "True"
	      | False => "False"
	      | Empty => "Empty"
	      | Box(f1) => "Box("^pp1(f1)^")"
	      | Diamond(f1) => "Diamond("^pp1(f1)^")"
	      | Neg(f1) => "Neg("^pp1(f1)^")"
	      | Chop(f1,f2) => "Chop("^pp1(f1)^","^pp1(f2)^")"
	      | And(f1,f2) => "And("^pp1(f1)^","^pp1(f2)^")"
	      | Or(f1,f2) => "Or("^pp1(f1)^","^pp1(f2)^")"
	      | Iff(f1,f2) => "Iff("^pp1(f1)^","^pp1(f2)^")"
	      | Imp(f1,f2) => "Imp("^pp1(f1)^","^pp1(f2)^")"
	      | Pred(s) => "Pred("^pps(s)^")"
	    and pps s =
		case s of
		    Id(st) => st
		  | tt => "tt"
		  | ff => "ff"
		  | sNeg(f1) => "sNeg("^pps(f1)^")"
		  | sAnd(f1,f2) => "sAnd("^pps(f1)^","^pps(f2)^")"
		  | sOr(f1,f2) => "sOr("^pps(f1)^","^pps(f2)^")"
		  | sIff(f1,f2) => "sIff("^pps(f1)^","^pps(f2)^")"
		  | sImp(f1,f2) => "sImp("^pps(f1)^","^pps(f2)^")"
	in pp1 f end
end
