(***************************************************************************
 File "natfun.sml" Updated 1993-02-10
 Duration Calculus checker, version 2
 Finite maps from the naturals, PS 1993-01-04
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature NATFUN = 
    sig 
	eqtype 'a natfun			
	exception AlreadyDefined
	exception Undefined
	val emptyfun: 'a natfun
	val singlefun: int * 'a -> 'a natfun
	val extend: 'a natfun * int * 'a -> 'a natfun   (* new not in domain *)
	val update: 'a natfun * int * 'a -> 'a natfun   (* may be in domain  *)
	val joinfun: 'a natfun * 'a natfun -> 'a natfun (* domains disjoint  *)
	val joinfuns: 'a natfun list -> 'a natfun       (* domains disjoint  *)
	val override: 'a natfun * 'a natfun -> 'a natfun
	val apply: 'a natfun -> int -> 'a
	val compose: 'a natfun * ('a -> 'b) -> 'b natfun
	val domain: 'a natfun -> int list
	val range: 'a natfun -> 'a list
	val mkassoclist: 'a natfun -> (int * 'a) list
	val mknatfun: (int * 'a) list -> 'a natfun
	val foldrange: ('a * 'b -> 'a) -> ('a * 'b natfun) -> 'a
    end

structure Natfun : NATFUN = 
struct 
    exception AlreadyDefined
    exception Undefined

    datatype 'a option = none | the of 'a

    datatype 'a natfun = 
	emptyfun 
      | some of 'a option * 'a natfun * 'a natfun

(* The value emptyfun is equivalent to some(none, emptyfun, emptyfun).
A natfun is *reduced* if it does not contain a subterm of the latter
form.  The functions above always produce reduced natfuns.  Reduced
natfuns are equal by "=" iff they represent the same function (and the
range is an equality type). *)

    fun sing (0, a) = some(the a, emptyfun, emptyfun)
      | sing (n, a) = if Bits.andb(n,1) <> 0 then
	                  some(none, sing(Bits.rshift(n,1), a), emptyfun)
		      else
			  some(none, emptyfun, sing(Bits.rshift(n,1) - 1, a))

    fun singlefun (i, a) = sing (i, a)
	
    fun mksome (none, emptyfun, emptyfun) = emptyfun
      | mksome (a,    t1,       t2      ) = some(a, t1, t2)

    fun joinfun (emptyfun, nf2) = nf2
      | joinfun (nf1, emptyfun) = nf1
      | joinfun (some(a1, t11, t12),   some(none, t21, t22)) =
	mksome(a1, joinfun(t11, t21), joinfun(t12, t22))
      | joinfun (some(none, t11, t12), some(a2, t21, t22)) =
	mksome(a2, joinfun(t11, t21), joinfun(t12, t22))
      | joinfun (some(the _, _, _), some(the _, _, _)) =
	raise AlreadyDefined

    fun joinfuns [] = emptyfun
      | joinfuns (nf::nfr) = joinfun(nf, joinfuns nfr)

    fun override (emptyfun, nf2) = nf2
      | override (nf1, emptyfun) = nf1
      | override (some(a1, t11, t12),   some(none, t21, t22)) =
	mksome(a1, override(t11, t21), override(t12, t22))
      | override (some(none, t11, t12), some(a2, t21, t22)) =
	mksome(a2, override(t11, t21), override(t12, t22))
      | override (some(_, t11, t12),    some(a2, t21, t22)) =
	mksome(a2, override(t11, t21), override(t12, t22))
	
    fun extend (nf, i, a) = joinfun(nf, singlefun(i,a))

    fun mknatfun []              = emptyfun
      | mknatfun ((n,a) :: rest) = extend(mknatfun rest, n, a)

    fun update (nf, i, a) = override(nf, singlefun(i,a))

    fun apply emptyfun              _ = raise Undefined
      | apply (some(none,  t1, t2)) 0 = raise Undefined
      | apply (some(the a, t1, t2)) 0 = a
      | apply (some(_,     t1, t2)) n = 
	if Bits.andb(n,1) <> 0 then
	    apply t1 (Bits.rshift(n,1))
	else
	    apply t2 (Bits.rshift(n,1) - 1)

    fun compose (emptyfun,            g) = emptyfun
      | compose (some(none,  t1, t2), g) = 
	some(none, compose(t1, g), compose(t2, g))
      | compose (some(the a, t1, t2), g) = 
	some(the (g a), compose(t1, g), compose(t2, g))

    (* mklist: (int * 'a -> 'b) -> 'a natfun -> 'b list *)

    fun mklist sel nf =
	let fun h (n, d, emptyfun,                    nl) = nl
	      | h (n, d, some(v, emptyfun, emptyfun), nl) = 
		(case v of none => nl | the a => sel(n,a) :: nl)
	      | h (n, d, some(v, t1,       emptyfun), nl) = 
		h(n+d, 2*d, t1, 
		  (case v of none => nl | the a => sel(n,a) :: nl))
	      | h (n, d, some(v, t1,       t2),       nl) = 
		h(n+d, 2*d, t1, 
		   h(n+2*d, 2*d, t2, 
		      (case v of none => nl | the a => sel(n,a) :: nl)))
	in h(0, 1, nf, []) end

    val domain = mklist (fn (n,a) => n) 
    val range  = mklist (fn (n,a) => a)
    val mkassoclist = mklist (fn x => x)

    (* foldrange: ('a * 'b -> 'a) -> ('a, 'b natfun) -> 'a *)

    fun foldrange f (e, nf) =
	let fun h (e,                    emptyfun) = e
	      | h (e, some(v, emptyfun, emptyfun)) = 
		(case v of none => e | the a => f(e,a))
	      | h (e, some(v, t1,       emptyfun)) = 
		h((case v of none => e | the a => f(e,a)), t1)
	      | h (e, some(v, t1,       t2)      ) = 
		h(h((case v of none => e | the a => f(e,a)),t2), t1)
	in h(e, nf) end

end

(***********************************************************************

Representation:

A finite map from the naturals (non-negative integers) is represented
by an applicative array, represented as a binary tree (as
in L.C. Paulson, "ML for the Working Programmer", pp. 130-132).
However, indexing in the tree is based on the dyadic rather than
binary representation of naturals. 

Complexity of the operations:

The operation emptyfun takes constant time; the operations
singlefun(i,a), extend(nf,i,a), update(nf,i,a) and (apply nf i) take
time proportional to log2(i); and the operations joinfun(nf1,nf2) and
override(nf1,nf2) take time proportional to
max(min(card(domain(nf1)),card(domain(nf2))), log2(max(domain(nf1) U
domain(nf2)))).  The operations compose(nf,f), range(nf) and
domain(nf) take time proportional to card(domain(nf)).

***********************************************************************)
