(***************************************************************************
 File "newnatsetset.sml" Updated 1993-02-12
 Duration Calculus checker, version 2 r 1
 Representation of finite sets of naturals, and finite functions on those
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature NATSETSET = 
    sig 
	eqtype natset			(* sets of non-negative integers *)
	eqtype 'a natsetfun
	exception Natsetapply
	val empty : natset
	val emptyset : 'a natsetfun
	val singleton : int -> natset
	val union : natset * natset -> natset
	val add : natset * int -> natset
	val intersection : natset * natset -> natset
	val cardinality : natset -> int
	val member : int * natset -> bool
	val members : natset -> int list
	val difference : natset * natset -> natset
	val delete : natset * int -> natset
	val foldset: ('a * int -> 'a) -> 'a * natset -> 'a
	val mapset: (int -> 'a) -> natset -> 'a list
	val disjoint : natset * natset -> bool
	val natsetof : int list -> natset
	val setfilter : (int -> bool) -> natset -> natset
	val fromto: int * int -> natset
	val indomain : natset * 'a natsetfun -> bool 
	val natsetapply: 'a natsetfun -> natset -> 'a	    
	val extendnatsetfun: 'a natsetfun * natset * 'a -> 'a natsetfun
    end

structure Natsetset: NATSETSET = 
struct 
    (* natsets: finite sets of naturals, PS 1992 *)
    datatype natset = 
	empty 
      | some of bool * natset * natset

    fun sing 0 = some(true, empty, empty)
      | sing n = if Bits.andb(n,1) <> 0 then
	             some(false, sing(Bits.rshift(n,1)), empty)
		 else
		     some(false, empty, sing(Bits.rshift(n,1) - 1))

    fun singleton i = sing i
	
    fun cardinality empty             = 0
      | cardinality (some(b, t1, t2)) = 
	(if b then 1 else 0) + cardinality t1 + cardinality t2

    fun mksome (false, empty, empty) = empty
      | mksome (b, t1, t2) = some(b, t1, t2)
	
    fun union (empty, ns2) = ns2
      | union (ns1, empty) = ns1
      | union (some(b1, t11, t12), some(b2, t21, t22)) =
	some(b1 orelse b2, union(t11, t21), union(t12, t22))
	
    fun add (is,i) = union(is, singleton i)

    fun intersection (empty, ns2) = empty
      | intersection (ns1, empty) = empty
      | intersection (some(b1, t11, t12), some(b2, t21, t22)) =
	mksome(b1 andalso b2, intersection(t11, t21), intersection(t12, t22))
	
    fun difference (empty, ns2) = empty
      | difference (ns1, empty) = ns1
      | difference (some(b1, t11, t12), some(b2, t21, t22)) =
	mksome(b1 andalso (not b2), difference(t11, t21), difference(t12, t22))
		  
    fun delete (is, i) = difference(is, singleton i)

    fun disjoint (empty, ns2) = true
      | disjoint (ns1, empty) = true
      | disjoint (some(b1, t11, t12), some(b2, t21, t22)) =
	not (b1 andalso b2) 
	andalso disjoint(t11, t21) 
	andalso disjoint (t12, t22)  

    fun member (i, is) = not(disjoint(is, singleton i))

    fun foldset f (e, t) =
	let fun sl (n, d, empty, a)                 = a
	      | sl (n, d, some(b, empty, empty), a) = 
		if b then f(a,n) else a
	      | sl (n, d, some(b, t1,    empty), a) = 
		sl(n+d, 2*d, t1, if b then f(a,n) else a)
	      | sl (n, d, some(b, t1,    t2), a)    = 
		sl(n+d, 2*d, t1, 
		   sl(n+2*d, 2*d, t2, if b then f(a,n) else a))
	in sl(0, 1, t, e) end

    fun mapset f t = foldset (fn (a,i) => f i :: a) ([], t)

    fun members t = foldset (fn (a,i) => i :: a) ([], t)

    fun natsetof []      = empty
      | natsetof (x::xs) = add(natsetof xs, x)

    fun fromto(i, j) = if i > j then empty else add(fromto(i+1, j), i)

    fun setfilter p t =
        let fun h (n, d, empty)           = empty
	      | h (n, d, some(b, t1, t2)) = 
		mksome(b andalso p n, h(n+d, 2*d, t1), h(n+2*d, 2*d, t2))
	in h(0, 1, t) end

(* natsetfuns: functions from sets of nonnegative integers, PS 1993-02-11 *)

    datatype 'a natsetfun = 
	Empty   				 (* domain = {}         *)
      | Sing of 'a				 (* domain = {empty}    *)
      | None of 'a natsetfun * 'a natsetfun      (* empty not in domain *)
      | Some of 'a * 'a natsetfun * 'a natsetfun (* empty in domain     *)

    (* Identities: Empty  = None(Empty,     Empty)
                   Sing a = Some(a, Sing a, Empty) *)

    val emptyset = Empty

(* val jnsf: 'a natsetfun * 'a natsetfun -> 'a natsetfun *) 

    fun jnsf (Empty, T2) = T2
      | jnsf (T1, Empty) = T1
      | jnsf (T1 as Sing _, Sing _) = T1 (* ?? *)
      | jnsf (T1 as Sing a, None(T21, T22)) = 
        Some(a, jnsf(T1, T21), T22)
      | jnsf (T1 as Sing a, Some(_, T21, T22)) =
	Some(a, jnsf(T1, T21), T22)
      | jnsf (None(T11, T12), T2 as Sing b) =
	Some(b, jnsf(T11, T2), T12)
      | jnsf (None(T11, T12), None(T21, T22)) = 
	None(jnsf(T11,T21), jnsf(T12, T22))
      | jnsf (None(T11, T12), Some(b, T21, T22)) = 
	Some(b, jnsf(T11, T21), jnsf(T12, T22))
      | jnsf (T1 as Some(a, T11, T12), T2 as Sing _) = 
	T1 (* Some(a, jnsf(T11, T2), T12)	  *) (* ?? *) 
      | jnsf (Some(a, T11, T12), None(T21, T22)) = 
	Some(a, jnsf(T11, T21), jnsf(T12, T22))
      | jnsf (Some(a, T11, T12), Some(_, T21, T22)) = 
	Some(a, jnsf(T11, T21), jnsf(T12, T22)) (* ?? *)

    fun depth empty = 0
      | depth (some(b, t1, t2)) = 1 + max(depth t1, depth t2)
	    
    fun singlensf(ns, a) =
	let (* The T argument of h is never Empty *)
	    fun h (0, empty, T as Sing _ )             = T
              | h (0, empty, T)                        = None(T, Empty)
              | h (0, some(false, _, _), T as Sing _ ) = T
	      | h (0, some(false, _, _), T)            = None(T, Empty)
	      | h (0, some(true,  _, _), T)            = None(Empty, T)
	      | h (n, empty, T) = 
		h(n-1, empty, h(n-1, empty, T))
	      | h (n, some(b, t1, t2), T) = 
                h(n-1, t1, h(n-1, t2, T))
	    fun rep (n, T) =
		if n < 0 then T else rep(n-1, h(n, ns, T))
	in rep (depth ns, Sing a) end

    fun extendnatsetfun (nss, ns, a) = jnsf(nss, singlensf(ns, a))

    fun get (ns, nss) =
        let fun h (0, empty, None(T1, T2)   ) = T1
              | h (0, empty, Some(_, T1, T2)) = T1
	      | h (0, empty, T)               = T 
	      | h (0, some(_, _, _), Empty)   = Empty
	      | h (0, some(b, _, _), T as Sing _ ) =
                if b then Empty else T
	      | h (0, some(b, _, _), None(T1, T2)) =
		if b then T2 else T1
	      | h (0, some(b, _, _), Some(_, T1, T2)) =
		if b then T2 else T1
	      | h (n, _, Empty)               = Empty
              | h (n, empty, T as Sing _ )    = T
	      | h (n, empty, T)               = h(n-1, empty, h(n-1, empty, T))
	      | h (n, some(b, t1, t2), T)     = h(n-1, t2, h(n-1, t1, T))
	    fun rep (n, T) =
		if n < 0 then T else h(n, ns, rep(n-1, T))
	in rep(depth ns, nss) end

    fun indomain (ns, nss) = 
        case get(ns, nss) of
	    Empty  => false
	  | Sing _ => true
          | None _ => false
	  | Some _ => true

    exception Natsetapply

    fun natsetapply nss ns =
	case get(ns, nss) of
            Empty         => raise Natsetapply
	  | Sing a        => a
	  | None _        => raise Natsetapply
	  | Some(a, _, _) => a
end
