(***************************************************************************
 File "nfa.sml" Updated 1993-02-12
 Duration Calculus checker, version 2 r 1
 Generating and running NFA from a DDC formula in simple form
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature NFA =
sig
    eqtype natset
    eqtype 'a natfun
    eqtype 'a Bdd
    type Formula
    type 'a option
    val denseDFAofneg: Formula * string list 
                       -> int Bdd natfun * int * natset * int
    val discreteDFAofneg: Formula * string list 
                       -> int Bdd natfun * int * natset * int
    val accepting: int Bdd natfun * int * natset * string list 
                   -> string list option
    val denseNFA: Formula * string list 
                  -> natset Bdd natfun * natset * natset * int
    val discreteNFA: Formula * string list 
                   -> natset Bdd natfun * natset * natset * int
    val NFA2DFA: (natset Bdd natfun * natset * natset * int) * int
                 -> int Bdd natfun * int * natset * int
end

functor Nfa(structure Natsetset: NATSETSET
	    and Natfun: NATFUN
	    and Bdd: BDD
	    and Formula: FORMULA
	    and Partition: PARTITION
	    sharing type Bdd.Bdd = Partition.Bdd
	    sharing type Natsetset.natset = Partition.natset
	    sharing type Natfun.natfun = Partition.natfun): NFA =
struct
    open Basic Natsetset Natfun Bdd Formula

    (* join(t1,t2) is the transition BDD from state {a1,a2}, when ti is 
     the transition bdd from state ai, i=1,2. *)

    val join = combinebdd union
    
    (* Construct the joint transition bdd for the set ns: natset of states. *)
	
    fun jointbdd (tra: natset Bdd natfun, ns: natset) = 
	let fun joinlist (bdd :: bdds) = foldright join (bdds, bdd)
	in reduce (joinlist (mapset (apply tra) ns)) end

(**************************************************************
           C O N S T R U C T I N G   A N   N F A
***************************************************************

When type nfatra = natset Bdd natfun, then

NFA: (nfatra*natset*nfatra*natset -> nfatra) * Formula * string list * int 
     -> nfatra * natset * natset * int

Function NFA maps (comb, f, snames, n) to (tra, ini, fin, n'), where
comb is a (discrete-time or dense-time) concatenator of automata, f is
a formula, snames is the list of state names, and n is the first
unused automaton state number; and tra is the transition relation, ini
the set of initial states, fin the set of final or accepting states,
and n' is the new least unused automaton state number. *)

fun NFA(comb,f:Formula,snames,n:int) =
    case f of
        Pred(s)     => NFA_Pred(s,snames,n)
      | Chop(f1,f2) => 
	    let val (tra1,ini1,fin1,n1) = NFA(comb,f1,snames,n)
		val (tra2,ini2,fin2,n2) = NFA(comb,f2,snames,n1)
		val tra1' = comb(tra1, fin1, tra2, ini2)
		val tra12 = joinfun(tra1', tra2)
		val ti2 = jointbdd(tra12, ini2)
		fun extd (tr,i) = 
		    update(tr, i, reduce(join(apply tra1 i, ti2)))
		val tra12' = foldset extd (tra12, fin1) 
		val fin12 = 
		    if disjoint(ini2,fin2) then fin2 else union(fin2, fin1)
	    in (tra12', ini1, fin12, n2) end
      | Or(f1,f2)   => 
	    let val (tra1,ini1,fin1,n1) = NFA(comb,f1,snames,n)
		val (tra2,ini2,fin2,n2) = NFA(comb,f2,snames,n1)
		val tra12 = joinfun(tra1,tra2)
	    in (tra12, union(ini1,ini2), union(fin1,fin2), n2) end
      | Neg(Neg(f1))=> DFA2NFA(NFA2DFA(NFA(comb,f1,snames,n),n))
      | Neg(f1)     => 
	    let val (tra,ini,fin,n1) = 
		DFA2NFA(NFA2DFA(NFA(comb,f1,snames,n),n))
		val fin1 = difference(fromto(n,n1-1),fin)
	    in  (tra,ini,fin1,n1) end

(* Function NFA_Pred constructs a three state NFA (corresponding to
'p') from a state expression p. *)

and NFA_Pred(s:State,snames,n:int) =
    let val init  = n			(* initial state *)
	val final = n+2			(* final state   *)
	val dead  = n+1			(* dead state    *)
	val finalset = singleton final
	val deadset  = singleton dead
	val bc = map sortstr (Formula.basic_conjuncts s)
        val sbdd    = buildbdd(snames,bc,deadset,finalset)
	val truebdd = buildbdd(snames,[[]],deadset,deadset)
	val trans = mknatfun[(init, sbdd),(final, sbdd),(dead, truebdd)]
    in (trans, singleton(init), finalset, n+3) end

(***************************************************************************
             C O N V E R T I N G   N F A   T O   D F A
***************************************************************************)
 
(* First find a triple
	(dstatefun, dtra, n'): int natsetfun * (natset * natset Bdd) list * n

of a renaming dstatefun of the states of the DFA, a list dtra of the
corresponding transition Bdds (counting from 0), and the number of new
states (equals the length of dtra).  Then rename the states from
natset to int, build a natfun instead of a list, and minimize the DFA.
Note that the state numbers of the intermediate DFA start at 0.  *)

and NFA2DFA ((tra, ini, fin, _), n) =
    let val (dstatefun, dtralist, n') = N2D ([ini], emptyset, [], 0, tra)
	val renamestate = natsetapply dstatefun
	fun mkdtra(dtra, (ns,nsbdd)) =
	    extend(dtra, renamestate ns, mapbdd renamestate nsbdd)
	val dtra = foldleft mkdtra (emptyfun, dtralist)
	val dstates = map (fn (ns, _) => ns) dtralist
	fun hasfinal ns = not (disjoint(ns, fin))
	val dfin = (natsetof o (map renamestate) o (filter hasfinal)) dstates
    in mindfa((dtra, renamestate ini, dfin, n'), n)  end

(* Given an NFA transition relation tra, find the set of states of the
corresponding DFA, and a map from dstates to integers.  Invariant: n =
length dtra *)

and N2D ([],        dstatefun, dtra, n, tra) = (dstatefun, dtra, n)
  | N2D (ns :: nsr, dstatefun, dtra, n, tra) =
    if indomain(ns, dstatefun) then N2D(nsr, dstatefun, dtra, n, tra)
    else let val nsbdd = jointbdd(tra, ns)
	     val reach = Bdd.range nsbdd
	 in N2D(reach @ nsr, extendnatsetfun(dstatefun, ns, n), 
		(ns, nsbdd) :: dtra, n+1, tra) 
	 end

(***************************************************************************
             M I N I M I Z I N G   T H E   D F A
***************************************************************************

Given a DFA (tra: int Bdd natfun, ini, fin, n), build the
minimized DFA (tra': int Bdd natfun, ini', fin', n').  First
compute a partition into consistent classes.  A class becomes a state
of the minimized DFA, after renaming: select a representative x from
the class, then classof x is the new name.  The set of initial states
init' of the new DFA is the set of renamed initial states, and
similarly for the set of final states. *)

and mindfa ((tra, ini, fin, n), next) = 
    let val (n', nss, classof)  = Partition.coarsestpartition(n, fin, tra)
	fun rename i = next + apply classof i
	fun extendtra (tr, x :: _) =
	    let val xbdd = reduce(mapbdd rename (apply tra x))
	    in extend(tr, rename x, xbdd) end
        val tra' = foldleft extendtra (emptyfun, nss)
	val ini' = rename ini
	val fin' = natsetof (mapset rename fin)
    in (tra', ini', fin', next + n') end

and DFA2NFA (tra, ini, fin, n) =
    (compose(tra, mapbdd singleton), singleton ini, fin, n)

fun DFAofneg (comb, f, snames) =
    let val (tra,ini,fin,n1) = NFA2DFA(NFA(comb,f,snames,0),0)
	val fin1 = difference(fromto(0,n1-1),fin)
    in  (tra,ini,fin1,n1) end

(* Combine automata A1 = (tra1, _, fin1) and A2 = (tra2, ini2, _), in
dense time, using contraction closure.

Let a be a state of A1, and let bdda be its BDD.  If some final state
of A1 is reachable from a, then (cc bdda) extends bdda with
transitions to every A2-state reachable (on the same letter) from the
joint initial state bddi2 of A2.  *)

fun densecombine (tra1, fin1, tra2, ini2)  = 
    let val bddi2 = jointbdd(tra2, ini2)	
	fun cc bdda = 
            let fun extendfin (bi,ba) =
		if not (disjoint(fin1, ba)) then union(bi,ba) else ba
	    in combinebdd extendfin (bddi2, bdda) end
    in compose(tra1, cc) end

fun denseNFA (f, snames) = NFA(densecombine, f, snames, 0)
    
fun denseDFAofneg (f, snames) = DFAofneg (densecombine, f, snames) 

fun discretecombine (tra1, _, _, _) = tra1

fun discreteNFA (f, snames) = NFA(discretecombine, f, snames, 0)
fun discreteDFAofneg (f, snames) = DFAofneg (discretecombine, f, snames) 

(* Search DFA for path to final state *)

fun accepting(tra, ini, fin, snames) =
    let fun newbranches i = Bdd.bdd2assoc (apply tra i, snames)
	fun h ([], path, [], marked) = None 
	  | h ([], _, (ts,newpath)::trr, marked) = 
	    h(ts, newpath, trr, marked)
	  | h ((p,i)::tr, path, tss, marked) = 
	    if member(i, marked)   then h(tr, path, tss, marked)
	    else if member(i, fin) then Some (rev (p::path))
	    else h(newbranches i, p :: path, (tr, path) :: tss, add(marked, i))
    in h (newbranches ini, [], [], empty) end
end
