(***************************************************************************
 File "partition.sml" Updated 1993-02-08
 Duration Calculus checker, version 2
 Computation of the coarsest consistent partition of DFA states
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

signature PARTITION =
sig
    eqtype natset
    eqtype 'a natfun
    eqtype 'a Bdd

    val coarsestpartition: int * natset * int Bdd natfun
        -> int * int list list * int natfun
end

functor Partition (structure Basic: BASIC
		   and Natsetset: NATSETSET
		   and Natfun: NATFUN
		   and Bdd: BDD) : PARTITION =
struct
    open Basic Natsetset Natfun Bdd

(* A partition is a triple (n, nss, classof): int * int list list * int natfun,
where 	n 	is the number of classes
	nss 	is a list of classes
	classof maps a state to the number of the class it belongs to.
Invariants: for all ns1, ns2, i, j, and c:
	n = length nss
	if ns1, ns2 in nss     then  ns1 and ns2 are disjoint
	classof i = classof j  iff   there is ns in nss such that i, j in ns
	if classof i = c       then  0 <= c < n

Definitions: Two int Bdds b1 and b2 are *equivalent* (wrt. classof) if
for corresponding pairs of leaves s1 and s2, classof s1 = classof s2.
A class ns is *consistent* (wrt. the partition and a transition
function) if for all states i, j in ns, the transition bdds of i and j
are equivalent (wrt. the partition).  A partition p is *consistent*
(wrt. a transition function) if every class in p is consistent wrt. p.

firstpartition: int * natset -> int * int list list * natfun
creates the first partition (into final resp. nonfinal states).  Let
n: int be the number of DFA states, and final: natset the set of
accepting states. *)

fun firstpartition (n: int, final: natset) =
    let fun h (0, [],  [],     classof) = (0, [],           emptyfun)
	  | h (0, fin, [],     classof) = (1, [fin],        classof)
	  | h (0, [],  nonfin, classof) = 
	    (1, [nonfin], compose(classof,fn x=>0))
	  | h (0, fin, nonfin, classof) = (2, [fin,nonfin], classof)
	  | h (i, fin, nonfin, classof) = 
	    let val i' = i-1 
	    in if member(i', final) then
		h(i', i'::fin, nonfin,     extend(classof, i', 0))
	    else
		h(i', fin,     i'::nonfin, extend(classof, i', 1))
	    end
    in h(n, [], [], emptyfun) end

(* Refine the partition into a consistent one.  Let p be a partition
and tra: int Bdd natfun the transition function of the DFA.  Then
refine(p, tra) is the coarsest consistent partition.  It is computed
by repeatedly checking that every class is consistent, and splitting
classes which are not consistent. *)

fun refine(p as (n, nss, classof), tra: int Bdd natfun) = 
    let fun split (x :: xr) =
	    let val xbdd     = apply tra x
		fun equivx y = comparebdd (apply classof) (xbdd, apply tra y)
		fun h ([],    t, f) = (t, f)
		  | h (y::yr, t, f) = if equivx y then h(yr, y::t, f   ) 
				                  else h(yr, t,    y::f)
	    in h(xr, [x], []) end
	fun newclass(n', neq) = 
	    foldleft (fn(co,i) => update(co,i,n')) (classof, neq)
	fun refi ([],      _   ) = p
	  | refi (ns::nsr, chkd) =
	    (case split ns of
		 (_,  [])  => refi(nsr, ns::chkd)
	       | (eq, neq) => (n+1, neq::eq::nsr@chkd, newclass(n, neq)))
	val p' as (n', _, _) = refi(nss, [])
    in if n = n' then p else refine(p',tra) end

fun coarsestpartition (n, final, tra) = 
    refine(firstpartition(n, final), tra)

end
