(***************************************************************************
 File "rdcdiag.sml" Updated 1993-02-10
 Duration Calculus checker, version 2
 Semantics of diagram language (DL) in a subset of Duration Calculus
 92-12-06 Jens Ulrik Skakkebaek
 Sestoft and Skakkebaek, Computer Science, Technical University of Denmark
***************************************************************************)

(*  A program written in the diagram language consists of:

    - Variable Declaration: (v,vals) list of variables v (string) and their 
      possible values (list of integers)

    - Initial Condition: Starting positions of the automaton (list of position
      numbers, i.e. integers)

    - Automaton: A list of transitions 

               -----                -----
              /     \   c -> va    /     \
              |  i  |------------->|  j  |
              \     /    [L,U]     \     /
               -----                -----
4
      where each transition is identified by

         1) From state i (integer),
         2) To state j (integer),
	 3) Enabling condition c (string), which must be of the form "x = k",
	 4) Assignment of variable va upon firing the transition (form: "x := k"),
	 5) Lower bound L on an enabled transition (integer),
	 6) Upper bound U on an enabled transition (integer). ~1 indicates inf.

******************************************************************************)

open Basic

(*=============================================================================
                 Definition of operator shorthands
=============================================================================*)

fun pred s = "'"^s^"'"
fun at i = "atl"^makestring(i:int)
fun at' i = pred ("atl"^makestring(i:int))
fun nat i = "(~atl"^makestring(i:int)^")"
fun nat' i = pred ("~atl"^makestring(i:int))
val tr = "true"

fun chop (h::t) = "("^h^(implode(map (fn x => (" ; "^x)) t))^")"
fun (*conj []     = "true"
  | *)conj (h::t) = "("^h^(implode(map (fn x => (" & "^x)) t))^")"
fun (*disj []     = "false"
  | *)disj (h::t) = "("^h^(implode(map (fn x => (" | "^x)) t))^")"
fun impl l r    = "("^l^" => "^r^")"
fun box s       = "([]("^s^"))"
fun dia s       = "(<>("^s^"))"
fun neg s       = "(~("^s^"))"

(*=============================================================================
                 Definition of time constraints
=============================================================================*)

(* len = 1 *)

val len_eq_1 = "('tt' & (~('tt';'tt')))"

(* len = i *)

fun len_eq i = 
    let fun l i = if i > 0 then len_eq_1::(l (i-1)) else ["''"]
    in chop (l i) end

(* len > i *)

fun len_gt i = chop [(len_eq i),(pred "tt")]

(* len >= i *)

fun len_geq i = chop [(len_eq i),tr]

(* len <= i *)

fun len_leq i = "(~("^(len_gt i)^"))"


fun nodes a = map (fn (i,_,_,_,_,_) => i) a

(*=============================================================================
                 The follow-operator

    'atli'->'atlj'  ==  [](('atli';true) => ('atli' | ('atli';'atlj';true)))
=============================================================================*)
fun follows i js = 
    (box (impl (chop [(at' i),(tr)])
	      (disj [(at' i),(chop [(at' i),(disj (map (at') js)),(tr)])]))) ^"\n"


(*=============================================================================
                 Upper limit on enabled transition

  For each node i, enabling transition c and upper limit U:
    []('atli & c' => len <= U)
=============================================================================*)
fun upper i c U = 
    if U = ~1 then "true\n" else 
	(box (impl (pred (conj [(at i), c])) (len_leq U)))^"\n"


(*=============================================================================
                 Lower limit on enabled transition

  For each transition from i to j, lower limit L, enabling condition c and
  assignment e:
      [](('~atli';'atli';'atlj') => <>((len>=L & 'atli & c');'atlj & e'))
=============================================================================*)
fun lower c e L i j =
(*    if L = 0 then "true\n" else*)
    if L = 1 then ((box (impl (chop [at'(i),at'(j)]) 
	      (dia (chop [(pred (conj [c,at(i)])),
			  (pred(conj [e,(at j)]))]))))^"\n")
    else     ((box (impl (chop [nat'(i),at'(i),at'(j)]) 
	      (dia (chop [(conj [len_geq(L),(pred (conj [c,at(i)]))]),
			  (pred(conj [e,(at j)]))]))))^"\n")

fun lowers c e L i js = conj (map (lower c e L i) js)
    
(*=============================================================================
                 Timing constraints
=============================================================================*)
fun time a = conj (map (fn (i,j,c,e,L,U) => conj [(upper i c U),(lower c e L i j)]) a )

fun time' (i,j,c,e,L,U)  = conj [(upper i c U),(lower c e L i j)]

(*=============================================================================
                 Disjointness of node identifiers

   Given n nodes in diagram:
   []('tt' => ' atl1 & ~atl2 & ... & ~atln  |
                atl2 & ~atl1 & ... & ~atln  |
		...
		atln & ~atl1 & ... & ~atln-1 ')
=============================================================================*)
fun nchop f nf js i = (conj ((f i)::(map (fn x => nf(x)) (filter (fn x => i<>x) js))))^"\n"

fun excl f nf s = box (impl (pred "tt") (pred (disj (map (nchop f nf s) s))))

fun autexcl a = excl (at) (nat) (setof (nodes a))


(*=============================================================================
                 Flow of automaton

    For all transitions from i to j and for any node k without a leaving transition:
    []('atl1'->'atl2') & ... & []('atli'->'atlj') & ... &
    []('atlk'->'atlk')
=============================================================================*)
fun ser_end ns i = if mem i ns then "true\n"
                   else (follows i [i])

fun nexts i a = map (fn (_,j,_,_,_,_) => j) (filter (fn (i',_,_,_,_,_) => i = i') a)

fun ser a = conj ((setof (map (fn (i,_,_,_,_,_) => (follows i (nexts i a))) a))@
                 (map (ser_end (nodes a)) (nodes a)))

fun ser' a = conj ((setof (map (fn (i,j,c,e,L,U) => conj [(follows i (nexts i a)),time'(i,j,c,e,L,U)]) a))@
                 (map (ser_end (nodes a)) (nodes a)))


(*=============================================================================
                 Automaton
=============================================================================*)
fun aut a = conj [(ser a),(time a),(autexcl a)]

fun auts al = conj (map (aut) al)


(*=============================================================================
                 Initial conditions

  For any node i specified to be the initial position in the diagram and 
  for all j transitions leaving i:
  'tt' => (('atli';true) & ... & 
      ('atli';'atlj';true => ('atli' & true;('c' & len >= L));'atlj & e';true)
	   & ... )
=============================================================================*)
fun init_lower e c L i j =
    (impl (chop [(at' i),(at' j),(tr)])
          (chop [(conj [(at' i), chop [tr,conj[(pred c),(len_geq L)]]]),
		 (pred (conj [e,(at j)])), (tr)]))^"\n"

fun init (i,j,c,e,L,U) =
    conj [(chop [at'(i),(tr)]), (conj (map (init_lower e c L i) [j]))]

fun inits is al =
    let val al' = foldright (op@) (al,[])
	val a' = filter (fn (i,_,_,_,_,_) => mem i is) al'
    in "("^(impl (pred "tt") (conj (map (init) a')))^")" end

(*=============================================================================
                 Disjointness of variable assignments

    For each variable x, initial value j, and all possible (n) assignments x:=i to x:
   []('tt' => ' x=1 & ~x=2 & ... & ~x=n  |
                x=2 & ~x=1 & ... & ~x=n  |
		...
		x=n & ~x=1 & ... & ~x=n-1 ') &
       ('tt'=>'x=j';true)    
=============================================================================*)
fun cond v i =  (v^"eq"^makestring(i:int))
fun cond' v i =  pred(v^"eq"^makestring(i:int))
fun ncond v i =  ("(~("^v^"eq"^makestring(i:int)^"))")
fun ncond' v i =  pred("(~("^v^"eq"^makestring(i:int)^"))")

fun varinit v init = impl (pred "tt") (chop [(cond' v init),tr])

fun vars vs = conj (map (fn (v,init,vals) => 
			 conj [(excl (cond v) (ncond v) vals), varinit v init])
		        vs)

(*=============================================================================
                 Framing
  For every declared variable x and all possible values u, x:=u must only be
  executed in the appropriate transitions from i to j:
		 []('~x=u';'x=u'=> <>('atli';'atlj'|'atlk';'atll'|...))
=============================================================================*)

fun frame al (v,_,vals) =
    let val fl =  map (fn x => (x,(filter (fn (_,_,_,y,_,_) => (cond v x)=y) 
				   (foldright (op@) (al,[]))))) vals
	fun f (a,al') = 
	    let val a' = (map (fn (i,j,_,_,_,_) => (dia (chop [at'(i),at'(j)])))
			   al')
	    in box (impl (chop [(ncond' v a),(cond' v a)])
		    (if a'=[] then (neg tr) else (disj a'))) end
    in conj (map (f) fl) end

fun frames al vs = conj (map (frame al) vs)

fun frame' al (v,_,vals) =
    let val fl =  map (fn x => (x,(filter (fn (_,_,_,y,_,_) => ((cond v x)<>y) andalso (y<>"tt"))
				   (foldright (op@) (al,[]))))) vals
	fun f (a,al') = box (impl (conj [(chop [(cond' v a),tr]),
					 neg ((*dia*) (disj (map (fn (i,j,_,_,_,_) => (chop [at'(i),at'(j)]))
					       al')))])
			           (cond' v a))
    in conj (map (f) fl) end
	
fun frames' al vs = conj (map (frame' al) vs)


(* [](('x=u';true & ~('atli';'atlj')) => 'x=u') *)
 
fun frame'' al (v,_,vals) =
    let val fl =  map (fn x => (x,(filter (fn (_,_,_,y,_,_) => ((cond v x)<>y) andalso (y<>"tt"))
				   (foldright (op@) (al,[]))))) vals
	fun f (a,al') = 
	    conj (map (fn (i,j,_,_,_,_) => 
		       box (impl (conj [(chop [(cond' v a),tr]),neg(chop [at'(i),at'(j)])])
			    (cond' v a)))
		  al')
    in conj (map (f) fl) end
	
fun frames'' al vs = conj (map (frame'' al) vs)

(*=============================================================================
                 Meaning of automaton
=============================================================================*)
fun diag (vs,is,al) = conj [(vars vs),(frames al vs),(inits is al),(auts al)]

fun diag' (vs,is,al) = conj [(auts al),(frames al vs)(*,(inits is al),(vars vs)*)]
